import logging
import vector

from particle import Particle

logger = logging.getLogger(__name__)


B_PDG_IDS = {int(p.pdgid) for p in Particle.findall(pdg_name="b")}
T_PDG_IDS = {int(p.pdgid) for p in Particle.findall(pdg_name="t")}
TAU_PDG_IDS = {int(p.pdgid) for p in Particle.findall(pdg_name="tau")}


class MadMinerParticle(vector.MomentumObject4D):
    """ """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.charge = None
        self.pdgid = None
        self.spin = None
        self.tau_tag = False
        self.b_tag = False
        self.t_tag = False

    def set_tags(self, tau_tag=False, b_tag=False, t_tag=False):
        self.tau_tag = tau_tag
        self.b_tag = b_tag
        self.t_tag = t_tag

    def set_pdgid(self, pdgid):
        self.pdgid = int(pdgid)

        try:
            self.charge = Particle.from_pdgid(self.pdgid).charge
        except RuntimeError:
            self.charge = 0.0

        if self.pdgid in B_PDG_IDS:
            self.b_tag = True
        elif self.pdgid in T_PDG_IDS:
            self.t_tag = True
        elif self.pdgid in TAU_PDG_IDS:
            self.tau_tag = True

    def set_spin(self, spin):
        self.spin = spin

    def __iadd__(self, other):
        assert isinstance(other, self.__class__)
        super().__iadd__(other)
        self.charge = None if self.charge is None or other.charge is None else self.charge + other.charge
        self.pdgid = None
        self.spin = None
        self.tau_tag = self.tau_tag or other.tau_tag
        self.b_tag = self.b_tag or other.b_tag
        self.t_tag = self.t_tag or other.t_tag
        return self

    def __isub__(self, other):
        assert isinstance(other, self.__class__)
        super().__isub__(other)
        self.charge = None if self.charge is None or other.charge is None else self.charge - other.charge
        self.pdgid = None
        self.spin = None
        self.tau_tag = None
        self.b_tag = None
        self.t_tag = None
        self.tau_tag = self.tau_tag or other.tau_tag
        self.b_tag = self.b_tag or other.b_tag
        self.t_tag = self.t_tag or other.t_tag
        return self

    def __add__(self, other):
        assert isinstance(other, self.__class__)
        vec = super().__add__(other)
        vec.charge = None if self.charge is None or other.charge is None else self.charge + other.charge
        vec.pdgid = None
        vec.spin = None
        vec.tau_tag = self.tau_tag or other.tau_tag
        vec.b_tag = self.b_tag or other.b_tag
        vec.t_tag = self.t_tag or other.t_tag
        return vec

    def __sub__(self, other):
        assert isinstance(other, self.__class__)
        vec = super().__sub__(other)
        vec.charge = None if self.charge is None or other.charge is None else self.charge - other.charge
        vec.pdgid = None
        vec.spin = None
        vec.tau_tag = self.tau_tag or other.tau_tag
        vec.b_tag = self.b_tag or other.b_tag
        vec.t_tag = self.t_tag or other.t_tag
        return vec

    def boost(self, *args):
        vec = super().boost(*args)

        particle = MadMinerParticle.from_xyzt(vec.x, vec.y, vec.z, vec.t)
        particle.charge = self.charge
        particle.spin = self.spin
        particle.pdgid = self.pdgid
        particle.tau_tag = self.tau_tag
        particle.b_tag = self.b_tag
        particle.t_tag = self.t_tag

        return particle



import h5py
import logging
import numpy as np
import shutil

from collections import OrderedDict
from contextlib import suppress
from typing import Callable
from typing import Iterator
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from madminer.models import AnalysisParameter
from madminer.models import NuisanceParameter
from madminer.models import Benchmark
from madminer.models import FiniteDiffBenchmark
from madminer.models import Observable
from madminer.models import Systematic
from madminer.models import SystematicScale
from madminer.models import SystematicType
from madminer.models import SystematicValue


logger = logging.getLogger(__name__)


def load_madminer_settings(file_name: str, include_nuisance_benchmarks: bool) -> tuple:
    """
    Loads the complete set of Madminer settings from a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to load the settings from
    include_nuisance_benchmarks: bool
        Whether or not to filter out the nuisance benchmarks

    Returns
    -------
    analysis_params: OrderedDict
    benchmarks: OrderedDict
    benchmark_nuisance_flags: list
    morphing_components: numpy.ndarray
    morphing_matrix numpy.ndarray
    observables: OrderedDict
    num_samples: int
    systematics: OrderedDict
    ref_benchmark: str
    nuisance_params: OrderedDict
    num_signal_events: numpy.ndarray
    num_background_events: int
    fin_differences: dict
    fin_diff_epsilon: float
    """

    analysis_params = _load_analysis_params(file_name)
    nuisance_params = _load_nuisance_params(file_name)

    (
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
        benchmark_reference_flags,
    ) = _load_benchmarks(file_name)

    (
        morphing_components,
        morphing_matrix,
    ) = _load_morphing(file_name)

    (
        fin_diff_base_benchmark,
        fin_diff_shift_benchmark,
        fin_diff_epsilon,
    ) = _load_finite_diffs(file_name)

    (
        observable_names,
        observable_defs,
    ) = _load_observables(file_name)

    (
        num_signal_events,
        num_background_events,
    ) = _load_samples_summary(file_name)

    (
        sample_observations,
        _,
        _,
    ) = _load_samples(file_name)

    (
        syst_names,
        syst_types,
        syst_values,
        syst_scales,
    ) = _load_systematics(file_name)

    # Build benchmarks dictionary
    benchmarks = OrderedDict()
    for b_name, b_matrix, b_nuisance_flag in zip(benchmark_names, benchmark_values, benchmark_nuisance_flags):

        # Filter out the nuisance benchmarks
        if include_nuisance_benchmarks is False and b_nuisance_flag is True:
            continue

        benchmarks[b_name] = Benchmark.from_params(b_name, analysis_params.keys(), b_matrix)

    # Build observables dictionary
    observables = OrderedDict()
    for o_name, o_def in zip(observable_names, observable_defs):
        observables[o_name] = Observable(o_name, o_def)

    # Build systematics dictionary
    systematics = OrderedDict()
    for s_name, s_type, s_value, s_scale in zip(syst_names, syst_types, syst_values, syst_scales):
        s_type = SystematicType.from_str(s_type)
        s_scale = SystematicScale.from_str(s_scale)
        systematics[s_name] = Systematic(s_name, s_type, s_value, s_scale)

    # Build finite differences dictionary
    fin_differences = OrderedDict()
    for base_name, matrix in zip(fin_diff_base_benchmark, fin_diff_shift_benchmark):
        fin_differences[base_name] = FiniteDiffBenchmark.from_params(base_name, analysis_params.keys(), matrix)

    # Compute concrete values
    num_samples = len(sample_observations)
    ref_benchmark = [name for name, flag in zip(benchmark_names, benchmark_reference_flags) if flag]
    ref_benchmark = ref_benchmark[0] if len(ref_benchmark) > 0 else None

    return (
        analysis_params,
        benchmarks,
        benchmark_nuisance_flags,
        morphing_components,
        morphing_matrix,
        observables,
        num_samples,
        systematics,
        ref_benchmark,
        nuisance_params,
        num_signal_events,
        num_background_events,
        fin_differences,
        fin_diff_epsilon,
    )


def save_madminer_settings(
    file_name: str,
    file_override: bool,
    parameters: Dict[str, AnalysisParameter],
    benchmarks: Dict[str, Benchmark],
    morphing_components: np.ndarray = None,
    morphing_matrix: np.ndarray = None,
    systematics: Dict[str, Systematic] = None,
    finite_differences: Dict[str, FiniteDiffBenchmark] = None,
    finite_differences_epsilon: float = None,
) -> None:
    """
    Saves the complete set of Madminer settings into a HDF5 data file

    Parameters
    ----------
    file_name: str
    file_override: bool
    parameters: OrderedDict
    benchmarks: OrderedDict
    morphing_components: numpy.ndarray
    morphing_matrix: numpy.ndarray
    systematics: OrderedDict
    finite_differences: OrderedDict
    finite_differences_epsilon: float

    Returns
    -------
        None
    """

    # Unpack provided dictionaries
    benchmark_names = [b.name for b in benchmarks.values()]
    benchmark_values = [[val for val in b.values.values()] for b in benchmarks.values()]

    fin_diffs_base_benchmarks = [b.base_name for b in finite_differences.values()]
    fin_diffs_shift_benchmarks = [[shift_name for shift_name in b.shift_names.values()] for b in finite_differences.values()]

    systematics_names = [s.name for s in systematics.values()]
    systematics_types = [s.type.value for s in systematics.values()]
    systematics_scales = [s.scale.value for s in systematics.values()]
    systematics_values = [s.value for s in systematics.values()]

    # Save information within the HDF5 file
    _save_analysis_parameters(file_name, file_override, parameters)
    _save_benchmarks(file_name, file_override, benchmark_names, benchmark_values)
    _save_morphing(file_name, file_override, morphing_components, morphing_matrix)

    if len(finite_differences) > 0:
        _save_finite_diffs(
            file_name,
            file_override,
            fin_diffs_base_benchmarks,
            fin_diffs_shift_benchmarks,
            finite_differences_epsilon,
        )
    if len(systematics) > 0:
        _save_systematics(
            file_name,
            file_override,
            systematics_names,
            systematics_types,
            systematics_values,
            systematics_scales,
        )


def save_nuisance_setup(
    file_name: str,
    file_override: bool,
    nuisance_benchmarks: List[str],
    nuisance_parameters: Dict[str, NuisanceParameter],
    reference_benchmark: str,
    copy_from_path: str = None,
) -> None:
    """
    Saves the names of nuisance-defined parameters and benchmarks in a HDF5 data file

    Parameters
    ----------
    file_name: str
    file_override: bool
    nuisance_benchmarks: list
    nuisance_parameters: dict
    reference_benchmark: str
    copy_from_path: str

    Returns
    -------
        None
    """

    if copy_from_path is not None:
        with suppress(OSError):
            shutil.copyfile(copy_from_path, file_name)

    (
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
        _,
    ) = _load_benchmarks(file_name)

    (
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
    ) = _add_benchmarks_custom(
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
        nuisance_benchmarks,
    )

    # Compute intermediate values
    benchmark_reference_flags = [name == reference_benchmark for name in benchmark_names]

    _save_benchmarks(
        file_name,
        file_override,
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
        benchmark_reference_flags,
    )

    if len(nuisance_parameters) > 0:
        _save_nuisance_params(
            file_name,
            file_override,
            nuisance_parameters,
        )


def load_events(
    file_name: str,
    start_index: int = 0,
    final_index: int = None,
    batch_size: int = 100_000,
    benchmark_nuisance_flags: List[bool] = None,
    sampling_benchmark: np.ndarray = None,
    sampling_factors: np.ndarray = None,
    include_nuisance_params: bool = True,
    include_sampling_ids: bool = False,
) -> Iterator[Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]
]:
    """
    Loads generated events information from a HDF5 data file

    Parameters
    ----------
    file_name: str
    start_index: int
    final_index: int
    batch_size: int
    benchmark_nuisance_flags: list
    sampling_benchmark: numpy.ndarray
    sampling_factors: numpy.ndarray
    include_nuisance_params: bool
    include_sampling_ids: bool

    Returns
    -------

    """

    # Nuisance benchmarks filtering
    if include_nuisance_params is False and benchmark_nuisance_flags is None:
        logger.warning("Lack of nuisance flags to filter out nuisance benchmarks")
        logger.warning("Processing all weights")
        include_nuisance_params = True

    elif include_nuisance_params is False and benchmark_nuisance_flags is not None:
        benchmark_filter = np.logical_not(np.array(benchmark_nuisance_flags, dtype=bool))

    (
        observations,
        weights,
        sampling_ids,
    ) = _load_samples(file_name)

    num_samples = len(observations)

    if start_index is None:
        start_index = 0
    if final_index is None:
        final_index = num_samples
    if batch_size is None:
        batch_size = num_samples

    final_index = min(num_samples, final_index)
    actual_index = start_index

    while actual_index < final_index:
        batch_final_index = min(actual_index + batch_size, final_index)

        batch_observations = observations[actual_index : batch_final_index]
        batch_weights = weights[actual_index : batch_final_index]
        batch_sampling_ids = None

        if include_nuisance_params is False:
            batch_weights = batch_weights[:, benchmark_filter]

        if sampling_ids.size > 0:
            batch_sampling_ids = sampling_ids[actual_index : batch_final_index]

            # Only return data matching sampling_benchmark
            if sampling_benchmark is not None:
                cut = np.logical_or(
                    batch_sampling_ids == sampling_benchmark,
                    batch_sampling_ids < 0,
                )

                batch_observations = batch_observations[cut]
                batch_weights = batch_weights[cut]
                batch_sampling_ids = batch_sampling_ids[cut]

            # Rescale weights based on sampling
            elif sampling_factors is not None:
                k_factors = sampling_factors[batch_sampling_ids]
                batch_weights = batch_weights * k_factors[:, np.newaxis]

        if include_sampling_ids:
            yield batch_observations, batch_weights, batch_sampling_ids
        else:
            yield batch_observations, batch_weights

        actual_index += batch_size


def save_events(
    file_name: str,
    file_override: bool,
    observables: Dict[str, Observable],
    observations: dict,
    weights: dict,
    sampling_benchmarks: List[int] = None,
    num_signal_events: List[int] = None,
    num_background_events: int = None,
) -> None:
    """
    Saves generated events information into a HDF5 data file

    Parameters
    ----------
    file_name: str
    file_override: bool
    observables: dict
    observations: dict
    weights: dict
    sampling_benchmarks: list
    num_signal_events: list
    num_background_events: int

    Returns
    -------
        None
    """

    # Unpack provided dictionaries
    observable_names = [o.name for o in observables.values()]
    observable_defs = [o.val_expression for o in observables.values()]
    observations = [val for val in observations.values()]

    _save_observables(file_name, file_override, observable_names, observable_defs)

    if weights is None or observations is None:
        return

    benchmark_names, _, _, _ = _load_benchmarks(file_name)

    logger.debug("Weight names to save in event file: %s", weights.keys())
    logger.debug("Benchmark names to save in event file: %s", benchmark_names)
    sorted_weights = _get_sorted_weights(benchmark_names, weights)

    sample_observations = np.array(observations).T
    sample_weights = np.array(sorted_weights).T
    sampling_ids = np.array(sampling_benchmarks, dtype=int)

    _save_samples(file_name, file_override, sample_observations, sample_weights, sampling_ids)
    _save_samples_summary(file_name, file_override, num_signal_events, num_background_events)


def _add_benchmarks_custom(
    benchmark_names: List[str],
    benchmark_values: List[np.ndarray],
    benchmark_nuisance_flags: List[bool],
    new_benchmark_names: List[str],
) -> Tuple[
    List[str],
    List[np.ndarray],
    List[bool]
]:
    """
    Extend the lists of benchmark properties with new custom benchmarks

    Parameters
    ----------
    benchmark_names: list
        List of benchmark names
    benchmark_values: list
        List of benchmark values per parameter
    benchmark_nuisance_flags: list
        List of flags indicating whether the benchmarks are nuisance or not
    new_benchmark_names: list
        List of custom benchmark names to add

    Returns
    -------
    benchmark_names: list
        List of benchmark names
    benchmark_values: list
        List of benchmark values per parameter
    benchmark_nuisance_flags: list
        List of flags indicating whether the benchmarks are nuisance or not
    """

    if isinstance(benchmark_values, np.ndarray):
        benchmark_values = list(benchmark_values)

    # Sort new benchmarks by name
    for new_name in sorted(new_benchmark_names):
        if new_name in benchmark_names:
            logger.debug(f"Benchmark {new_name} already in the list of benchmarks")
            continue

        logger.debug(f"Adding custom benchmark: {new_name}")
        benchmark_names.append(new_name)
        benchmark_values.append(np.zeros_like(benchmark_values[0]))
        benchmark_nuisance_flags.append(True)

    return (
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
    )


def _load_benchmarks(file_name: str) -> Tuple[List[str], List[List[float]], List[bool], List[bool]]:
    """
    Load benchmark properties from a HDF5 data file.

    Parameters
    ----------
    file_name: str
        HDF5 file name to load benchmark properties from

    Returns
    -------
    benchmark_names: list
        List of benchmark names
    benchmark_values: list
        List of benchmark values per parameter
    benchmark_nuisance_flags: list
        List of flags indicating whether the benchmarks are nuisance or not
    benchmark_reference_flags: list
        List of flags indicating whether the benchmarks are reference or not
    """

    with h5py.File(file_name, "r") as file:

        # Mandatory properties
        try:
            benchmark_names = file["benchmarks/names"][()]
            benchmark_values = file["benchmarks/values"][()]
        except KeyError:
            raise IOError("Cannot read benchmarks from HDF5 file")
        else:
            benchmark_names = _decode_strings(benchmark_names)

        # Optional properties
        try:
            benchmark_nuisance_flags = file["benchmarks/is_nuisance"][()]
            benchmark_nuisance_flags = [bool(flag) for flag in benchmark_nuisance_flags]
        except KeyError:
            logger.info("HDF5 file does not contain benchmark nuisance flags")
            benchmark_nuisance_flags = [False for _ in benchmark_names]

        try:
            benchmark_reference_flags = file["benchmarks/is_reference"][()]
            benchmark_reference_flags = [bool(flag) for flag in benchmark_reference_flags]
        except KeyError:
            logger.info("HDF5 file does not contain benchmark reference flags")
            benchmark_reference_flags = [False for _ in benchmark_names]

    return (
        benchmark_names,
        benchmark_values,
        benchmark_nuisance_flags,
        benchmark_reference_flags,
    )


def _save_benchmarks(
    file_name: str,
    file_override: bool,
    benchmark_names: List[str],
    benchmark_values: List[List[float]],
    benchmark_nuisance_flags: List[bool] = None,
    benchmark_reference_flags: List[bool] = None,
) -> None:
    """
    Save benchmark properties into a HDF5 data file.

    Parameters
    ----------
    file_name: str
        HDF5 file name to save benchmark properties into
    file_override: bool
        Whether to override HDF5 file contents or not
    benchmark_names: list
        List of benchmark names
    benchmark_values: list
        List of benchmark values per parameter
    benchmark_nuisance_flags: list
        List of flags indicating whether the benchmarks are nuisance or not
    benchmark_reference_flags: list
        List of flags indicating whether the benchmarks are reference or not

    Returns
    -------
        None
    """

    if benchmark_nuisance_flags is None:
        benchmark_nuisance_flags = [False for _ in benchmark_names]

    if benchmark_reference_flags is None:
        benchmark_reference_flags = [False for _ in benchmark_names]

    benchmark_names = _encode_strings(benchmark_names)
    benchmark_nuisance_flags = [int(flag) for flag in benchmark_nuisance_flags]
    benchmark_reference_flags = [int(flag) for flag in benchmark_reference_flags]

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["benchmarks"]

        # Store benchmarks
        file.create_dataset("benchmarks/names", data=benchmark_names, dtype="S256")
        file.create_dataset("benchmarks/values", data=benchmark_values)
        file.create_dataset("benchmarks/is_nuisance", data=benchmark_nuisance_flags)
        file.create_dataset("benchmarks/is_reference", data=benchmark_reference_flags)


def _load_finite_diffs(file_name: str) -> Tuple[List[str], List[List[str]], float]:
    """
    Load finite differences between benchmarks from a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to load observables properties into

    Returns
    -------
    fin_diff_base_benchmarks: list
        List of base benchmark names
    fin_diff_shift_benchmarks: list
        List of shift benchmark names, for each of the parameters
    fin_diff_epsilon: float
        Value representing the magnitude of the numerical uncertainty
    """

    fin_diff_base_benchmarks = []
    fin_diff_shift_benchmarks = []
    fin_diff_epsilon = 0.0

    with h5py.File(file_name, "r") as file:
        try:
            fin_diff_base_benchmarks = file["finite_differences/base_benchmarks"][()]
            fin_diff_shift_benchmarks = file["finite_differences/shifted_benchmarks"][()]
            fin_diff_epsilon = float(file["finite_differences/epsilon"][()])
        except KeyError:
            logger.error("HDF5 file does not contain finite difference information")
        else:
            fin_diff_base_benchmarks = _decode_strings(fin_diff_base_benchmarks)
            fin_diff_shift_benchmarks = [_decode_strings(names) for names in fin_diff_shift_benchmarks]

    return (
        fin_diff_base_benchmarks,
        fin_diff_shift_benchmarks,
        fin_diff_epsilon,
    )


def _save_finite_diffs(
    file_name: str,
    file_override: bool,
    fin_diff_base_benchmarks: List[str],
    fin_diff_shift_benchmarks: List[List[str]],
    fin_diff_epsilon: float,
) -> None:
    """
    Save finite differences between benchmarks into a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to save finite difference benchmarks into
    file_override: bool
        Whether to override HDF5 file contents or not
    fin_diff_base_benchmarks: list
        List of base benchmark names
    fin_diff_shift_benchmarks: list
        List of shift benchmark names, for each of the parameters
    fin_diff_epsilon: float
        Value representing the magnitude of the numerical uncertainty

    Returns
    -------
        None
    """

    fin_diff_base_benchmarks = _encode_strings(fin_diff_base_benchmarks)
    fin_diff_shift_benchmarks = [_encode_strings(names) for names in fin_diff_shift_benchmarks]

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["finite_differences"]

        file.create_dataset("finite_differences/base_benchmarks", data=fin_diff_base_benchmarks)
        file.create_dataset("finite_differences/shifted_benchmarks", data=fin_diff_shift_benchmarks)
        file.create_dataset("finite_differences/epsilon", data=fin_diff_epsilon)


def _load_morphing(file_name: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Load morphing properties from a HDF5 data file

    Parameters
    ----------
    file_name : str
        HDF5 file name to load morphing properties from

    Returns
    -------
    morphing_components : numpy.ndarray
    morphing_matrix : numpy.ndarray
    """

    morphing_components = None
    morphing_matrix = None

    with h5py.File(file_name, "r") as file:
        try:
            morphing_components = file["morphing/components"][()]
            morphing_matrix = file["morphing/morphing_matrix"][()]
        except KeyError:
            logger.error("HDF5 file does not contain morphing information")
        else:
            morphing_components = np.asarray(morphing_components, dtype=int)
            morphing_matrix = np.asarray(morphing_matrix, dtype=float)

    return morphing_components, morphing_matrix


def _save_morphing(
    file_name: str,
    file_override: bool,
    morphing_components: np.ndarray,
    morphing_matrix: np.ndarray,
) -> None:
    """
    Save morphing properties into a HDF5 data file

    Parameters
    ----------
    file_name : str
    file_override : bool
    morphing_components : numpy.ndarray
    morphing_matrix : numpy.ndarray

    Returns
    -------
        None
    """

    if morphing_components is None or morphing_matrix is None:
        return

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["morphing"]

        file.create_dataset("morphing/components", data=morphing_components.astype(int))
        file.create_dataset("morphing/morphing_matrix", data=morphing_matrix.astype(float))


def _load_nuisance_params(file_name: str) -> Dict[str, NuisanceParameter]:
    """
    Load nuisance parameter properties from a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to load nuisance parameter properties from

    Returns
    -------
    parameters: dict
        Dictionary of nuisance parameter properties to each parameter name
    """

    # Replace by normal dictionary on python 3.7+
    parameters = OrderedDict()

    with h5py.File(file_name, "r") as file:
        try:
            param_names = file["nuisance_parameters/names"][()]
            param_systematics = file["nuisance_parameters/systematics"][()]
            param_benchmarks_pos = file["nuisance_parameters/benchmark_positive"][()]
            param_benchmarks_neg = file["nuisance_parameters/benchmark_negative"][()]
        except KeyError:
            logger.error("HDF5 file does not contain nuisance parameters information")
            return parameters
        else:
            param_names = _decode_strings(param_names)
            param_systematics = _decode_strings(param_systematics)
            param_benchmarks_pos = _decode_strings(param_benchmarks_pos)
            param_benchmarks_neg = _decode_strings(param_benchmarks_neg)
            param_benchmarks_pos = [None if name == "" else name for name in param_benchmarks_pos]
            param_benchmarks_neg = [None if name == "" else name for name in param_benchmarks_neg]

    for name, sys, benchmark_pos, benchmark_neg in zip(
        param_names,
        param_systematics,
        param_benchmarks_pos,
        param_benchmarks_neg,
    ):
        parameters[name] = NuisanceParameter(name, sys, benchmark_pos, benchmark_neg)

    # TODO: The dictionary has been preserved. Harmony with other loaders?

    return parameters


def _save_nuisance_params(
    file_name: str,
    file_override: bool,
    parameters: Dict[str, NuisanceParameter],
) -> None:
    """
    Save nuisance parameter properties into a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to save nuisance parameter properties into
    file_override: bool
        Whether to override HDF5 file contents or not
    parameters: dict
        Dictionary of nuisance parameter properties to each parameter name

    Returns
    -------
        None
    """

    param_names = [p.name for p in parameters.values()]
    param_systematics = [p.systematic for p in parameters.values()]
    param_benchmarks_pos = [p.benchmark_pos if p.benchmark_pos else "" for p in parameters.values()]
    param_benchmarks_neg = [p.benchmark_neg if p.benchmark_neg else "" for p in parameters.values()]

    param_names = _encode_strings(param_names)
    param_systematics = _encode_strings(param_systematics)
    param_benchmarks_pos = _encode_strings(param_benchmarks_pos)
    param_benchmarks_neg = _encode_strings(param_benchmarks_neg)

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["nuisance_parameters"]

        file.create_dataset(
            "nuisance_parameters/names",
            data=param_names,
            dtype="S256",
        )
        file.create_dataset(
            "nuisance_parameters/systematics",
            data=param_systematics,
            dtype="S256",
        )
        file.create_dataset(
            "nuisance_parameters/benchmark_positive",
            data=param_benchmarks_pos,
            dtype="S256",
        )
        file.create_dataset(
            "nuisance_parameters/benchmark_negative",
            data=param_benchmarks_neg,
            dtype="S256",
        )

        # TODO: The dictionary has been preserved. Harmony with other loaders?


def _load_analysis_params(file_name: str) -> Dict[str, AnalysisParameter]:
    """
    Load analysis parameter properties from a HDF5 data file

    Parameters
    ----------
    file_name : str
        HDF5 file name to load analysis parameter properties from

    Returns
    -------
    parameters : dict
        Dictionary of analysis parameter properties to each parameter name
    """

    # Replace by normal dictionary on python 3.7+
    parameters = OrderedDict()

    with h5py.File(file_name, "r") as file:
        try:
            param_names = file["parameters/names"][()]
            param_lha_blocks = file["parameters/lha_blocks"][()]
            param_lha_ids = file["parameters/lha_ids"][()]
            param_max_power = file["parameters/max_power"][()]
            param_val_ranges = file["parameters/val_ranges"][()]
            param_transforms = file["parameters/transforms"][()]
        except KeyError:
            raise IOError("Cannot read parameters from HDF5 file")
        else:
            param_names = _decode_strings(param_names)
            param_lha_blocks = _decode_strings(param_lha_blocks)
            param_transforms = _decode_strings(param_transforms)

    for name, block, id, max_power, range, transform in zip(
        param_names,
        param_lha_blocks,
        param_lha_ids,
        param_max_power,
        param_val_ranges,
        param_transforms,
    ):
        parameters[name] = AnalysisParameter(
            str(name),
            str(block),
            int(id),
            int(max_power),
            tuple(range),
            str(transform),
        )

    return parameters


def _save_analysis_parameters(
    file_name: str, file_override: bool, parameters: Dict[str, AnalysisParameter]
) -> None:
    """
    Save analysis parameter properties into a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to save analysis parameter properties into
    file_override: bool
        Whether to override HDF5 file contents or not
    parameters: dict
        Dictionary of analysis parameter properties to each parameter name

    Returns
    -------
        None
    """

    param_names = [p.name for p in parameters.values()]
    param_lha_blocks = [p.lha_block for p in parameters.values()]
    param_lha_ids = [p.lha_id for p in parameters.values()]
    param_max_power = [p.max_power for p in parameters.values()]
    param_val_ranges = [p.val_range for p in parameters.values()]
    param_transforms = [p.transform for p in parameters.values()]

    param_names = _encode_strings(param_names)
    param_lha_blocks = _encode_strings(param_lha_blocks)
    param_transforms = _encode_strings(param_transforms)

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["parameters"]

        file.create_dataset("parameters/names", data=param_names, dtype="S256")
        file.create_dataset("parameters/lha_blocks", data=param_lha_blocks, dtype="S256")
        file.create_dataset("parameters/lha_ids", data=param_lha_ids)
        file.create_dataset("parameters/max_power", data=param_max_power)
        file.create_dataset("parameters/val_ranges", data=param_val_ranges)
        file.create_dataset("parameters/transforms", data=param_transforms, dtype="S256")


def _load_observables(file_name: str) -> Tuple[List[str], List[str]]:
    """
    Load observable properties from a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to load observables properties from

    Returns
    -------
    observable_names: list
        List of observable names
    observable_defs: list
        List of observable string-encoded definitions
    """

    observable_names = []
    observable_defs = []

    with h5py.File(file_name, "r") as file:

        try:
            observable_names = file["observables/names"][()]
            observable_defs = file["observables/definitions"][()]
        except KeyError:
            logger.error("HDF5 file does not contain observables information")
        else:
            observable_names = _decode_strings(observable_names)
            observable_defs = _decode_strings(observable_defs)

    return observable_names, observable_defs


def _save_observables(
    file_name: str,
    file_override: bool,
    observable_names: List[str],
    observable_defs: List[Union[str, Callable]],
) -> None:
    """
    Save observable properties into a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to save observables properties into
    file_override: bool
        Whether to override HDF5 file contents or not
    observable_names : list
        List of observable names
    observable_defs: list
        List of observable string-encoded or callable function definitions

    Returns
    -------
        None
    """

    observable_names = _encode_strings(observable_names)

    # Filter out callable definitions when saving into HDF5 file
    observable_defs = [d if isinstance(d, str) else "" for d in observable_defs]
    observable_defs = _encode_strings(observable_defs)

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["observables"]

        file.create_dataset("observables/names", data=observable_names, dtype="S256")
        file.create_dataset("observables/definitions", data=observable_defs, dtype="S256")


def _load_samples(file_name: str) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Load sample properties from a HDF5 data file

    Parameters
    ----------
    file_name: str
        HDF5 file name to load sample properties from

    Returns
    -------
    sample_observations: numpy.ndarray
    sample_weights: numpy.ndarray
    sampling_ids: numpy.ndarray
    """

    sample_observations = np.asarray([])
    sample_weights = np.asarray([])
    sampling_ids = np.asarray([])

    with h5py.File(file_name, "r") as file:
        try:
            sample_observations = file["samples/observations"][()]
            sample_weights = file["samples/weights"][()]
            sampling_ids = file["samples/sampling_benchmarks"][()]
        except KeyError:
            logger.error("HDF5 file does not contain sample information")
        else:
            assert sample_observations.shape[0] == sample_weights.shape[0], \
                "The number of sample observations and sample weights do not match"

    return (
        sample_observations,
        sample_weights,
        sampling_ids,
    )


def _save_samples(
    file_name: str,
    file_override: bool,
    sample_observations: np.ndarray,
    sample_weights: np.ndarray,
    sampling_ids: np.ndarray,
) -> None:
    """
    Load sample properties into a HDF5 data file.

    Parameters
    ----------
    file_name: str
        HDF5 file name to save sample properties into
    file_override: bool
        Whether to override HDF5 file contents or not
    sample_observations: numpy.ndarray
    sample_weights: numpy.ndarray
    sampling_ids: numpy.ndarray

    Returns
    -------
        None
    """

    assert sample_observations is not None
    assert sample_weights is not None
    assert sampling_ids is not None

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["samples"]

        file.create_dataset("samples/observations", data=sample_observations)
        file.create_dataset("samples/weights", data=sample_weights)
        file.create_dataset("samples/sampling_benchmarks", data=sampling_ids)


def _load_samples_summary(file_name: str) -> Tuple[np.ndarray, int]:
    """
    Load the number of signal and background events

    Parameters
    ----------
    file_name: str
        HDF5 file name to load sample numbers from

    Returns
    -------
    num_signal_events: numpy.ndarray
        List of signal events number per benchmark
    num_background_events: int
        Number of background events
    """

    num_signal_events = np.asarray([])
    num_background_events = 0

    with h5py.File(file_name, "r") as file:
        try:
            num_signal_events = file["sample_summary/signal_events"][()]
            num_background_events = file["sample_summary/background_events"][()]
        except KeyError:
            logger.error("HDF5 file does not contain sample summary information")

    return (
        num_signal_events,
        num_background_events,
    )


def _save_samples_summary(
    file_name: str,
    file_override: bool,
    num_signal_events: List[int],
    num_background_events: int,
) -> None:
    """
    Save the number of signal and background events

    Parameters
    ----------
    file_name: str
        HDF5 file name to save sample numbers into
    file_override: bool
        Whether to override HDF5 file contents or not
    num_signal_events: list
        List with the number of signal events per benchmark
    num_background_events: int
        Number of background events

    Returns
    -------
        None
    """

    if num_signal_events is None:
        num_signal_events = []

    if num_background_events is None:
        num_background_events = 0

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["sample_summary"]

        file.create_dataset("sample_summary/signal_events", data=num_signal_events)
        file.create_dataset("sample_summary/background_events", data=num_background_events)


def _load_systematics(file_name: str) -> Tuple[List[str], List[str], List[SystematicValue], List[str]]:
    """
    Load systematics properties into a HDF5 data file.

    Parameters
    ----------
    file_name: str
        HDF5 file name to load systematics properties from

    Returns
    -------
    systematics_names: list
        List of systematics names
    systematics_types: list
        List of systematics types
    systematics_values: list
        List of systematics values (str or float)
    systematics_scales: list
        List of systematics scales
    """

    systematics_names = []
    systematics_types = []
    systematics_values = []
    systematics_scales = []

    with h5py.File(file_name, "r") as file:
        try:
            systematics_names = file["systematics/names"][()]
            systematics_types = file["systematics/types"][()]
            systematics_values = file["systematics/values"][()]
            systematics_scales = file["systematics/scales"][()]
        except KeyError:
            logger.error("HDF5 file does not contain systematic information")
        else:
            systematics_names = _decode_strings(systematics_names)
            systematics_types = _decode_strings(systematics_types)
            systematics_scales = _decode_strings(systematics_scales)
            systematics_scales = [None if scale == "" else scale for scale in systematics_scales]

    return (
        systematics_names,
        systematics_types,
        systematics_values,
        systematics_scales,
    )


def _save_systematics(
    file_name: str,
    file_override: bool,
    systematics_names: List[str],
    systematics_types: List[str],
    systematics_values: List[SystematicValue],
    systematics_scales: List[str],
) -> None:
    """
    Save systematics properties into a HDF5 data file.

    Parameters
    ----------
    file_name: str
        HDF5 file name to save systematics properties into
    file_override: bool
        Whether to override HDF5 file contents or not
    systematics_names: list
        List of systematics names
    systematics_types: list
        List of systematics types
    systematics_values: list
        List of systematics values (str or float)
    systematics_scales: list
        List of systematics scales

    Returns
    -------
        None
    """

    systematics_scales = [scale if scale else "" for scale in systematics_scales]

    systematics_names = _encode_strings(systematics_names)
    systematics_types = _encode_strings(systematics_types)
    systematics_scales = _encode_strings(systematics_scales)

    # Append if file exists, otherwise create
    with h5py.File(file_name, "a") as file:

        if file_override:
            with suppress(KeyError):
                del file["systematics"]

        file.create_dataset("systematics/names", data=systematics_names, dtype="S256")
        file.create_dataset("systematics/types", data=systematics_types, dtype="S256")
        file.create_dataset("systematics/values", data=systematics_values, dtype="S256")
        file.create_dataset("systematics/scales", data=systematics_scales, dtype="S256")


def _encode_strings(strings: List[str]) -> List[bytes]:
    """
    Encodes a list of strings as bytes

    Parameters
    ----------
    strings : list
        List of any-codification strings

    Returns
    -------
    strings: list
        List of ASCII encoded bytes
    """

    return [s.encode("ascii", "ignore") for s in strings]


def _decode_strings(strings: List[bytes]) -> List[str]:
    """
    Decodes a list of bytes into a list of strings

    Parameters
    ----------
    strings : list
        List of any-codification bytes

    Returns
    -------
    strings: list
        List of ASCII decoded strings
    """

    return [s.decode("ascii") for s in strings]


def _get_sorted_weights(benchmark_names: List[str], weights: Dict[str, float]) -> List[float]:
    """
    Extracts and sorts weight values from a dictionary.
    First the benchmarks in the right order, then the remaining ones alphabetically

    Parameters
    ----------
    benchmark_names: list
        List of benchmark names
    weights: dict
        Dictionary of benchmark-to-weight values

    Returns
    -------
    sorted_weights: list
        List of sorted weights
    """

    sorted_keys = sorted(weights.keys())

    try:
        benchmark_weights = [weights[name] for name in benchmark_names]
        remaining_weights = [weights[key] for key in sorted_keys if key not in benchmark_names]
    except KeyError as error:
        logger.warning("Issue matching weight names between the HepMC file and MadMiner:")
        logger.warning(error)
        sorted_weights = [weights[key] for key in weights]
    else:
        sorted_weights = [*benchmark_weights, *remaining_weights]

    return sorted_weights



from .__info__ import __authors__, __version__
from .analysis import DataAnalyzer
from .core import MadMiner
from .delphes import DelphesReader
from .fisherinformation import (
    FisherInformation,
    InformationGeometry,
    profile_information,
    project_information,
)
from .lhe import LHEReader
from .likelihood import (
    HistoLikelihood,
    NeuralLikelihood,
    fix_params,
    project_log_likelihood,
    profile_log_likelihood,
)
from .limits import AsymptoticLimits
from .ml import (
    ParameterizedRatioEstimator,
    DoubleParameterizedRatioEstimator,
    LikelihoodEstimator,
    ScoreEstimator,
    MorphingAwareRatioEstimator,
    Ensemble,
    load_estimator,
)
from .plotting import (
    plot_uncertainty,
    plot_systematics,
    plot_pvalue_limits,
    plot_distribution_of_information,
    plot_fisher_information_contours_2d,
    plot_fisherinfo_barplot,
    plot_nd_morphing_basis_slices,
    plot_nd_morphing_basis_scatter,
    plot_2d_morphing_basis,
    plot_histograms,
    plot_distributions,
)
from .sampling import (
    SampleAugmenter,
    combine_and_shuffle,
    benchmark,
    benchmarks,
    morphing_point,
    morphing_points,
    random_morphing_points,
    iid_nuisance_parameters,
    nominal_nuisance_parameters,
)


import logging

logging.getLogger("madminer").addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)

logger.info("")
logger.info("|{}|".format("-" * 74))
logger.info("|{}|".format(" " * 74))
logger.info("|{}|".format(" MadMiner v{}".format(__version__).ljust(74)))
logger.info("|{}|".format(" " * 74))
logger.info("|{}|".format(" {}".format(__authors__).ljust(74)))
logger.info("|{}|".format(" " * 74))
logger.info("|{}|".format("-" * 74))
logger.info("")



from dataclasses import dataclass
from enum import Enum
from typing import Union


SystematicValue = Union[str, float]


class SystematicType(str, Enum):
    """
    Type of the Systematic value:

    - "pdf": the nuisance parameters effect will be determined by varying the PDF used.
    - "norm": it will affect the overall normalization of one or multiple samples in the process.
    - "scale": the nuisance parameter effect will be determined by varying factorization or regularization scales.
    """

    PDF = "pdf"
    NORM = "norm"
    SCALE = "scale"

    @classmethod
    def from_str(cls, value):
        for k, v in cls.__members__.items():
            if v == value:
                return cls[k]

        raise ValueError(f"Invalid systematic type: {value}")


class SystematicScale(str, Enum):
    """
    Scale of Systematic value:

    - "mu": sets both factorization and regularization scales vary
    - "muf": sets that only the factorization scale varies
    - "mur": sets that only the regularization scale varies
    """

    MU = "mu"
    MUF = "muf"
    MUR = "mur"

    @classmethod
    def from_str(cls, value):
        for k, v in cls.__members__.items():
            if v == value:
                return cls[k]

        raise ValueError(f"Invalid systematic scale: {value}")


@dataclass
class Systematic:

    name: str
    type: SystematicType
    value: SystematicValue
    scale: SystematicScale = None

    def __str__(self) -> str:
        """
        Formats the systematic in a nice way

        Returns
        -------
            Formatted systematic string
        """

        return (
            f"{self.name} "
            f"("
            f"Type: {self.type}, "
            f"Value: {self.value}, "
            f"Scale: {self.scale}"
            f")"
        )

    def __post_init__(self):
        """Perform certain attribute quality assertions"""

        if self.type != SystematicType.SCALE:
            assert self.scale is None, "Cannot specify scale"

        if self.type == SystematicType.SCALE:
            assert self.scale is not None, "Must specify scale"



from dataclasses import dataclass
from contextlib import suppress
from typing import Callable
from typing import Union


@dataclass
class Cut:

    name: str
    val_expression: str
    is_required: bool = False

    def __post_init__(self):
        """Perform certain attribute quality assertions"""

        with suppress(NameError):
            eval(self.val_expression)


@dataclass
class Efficiency:

    name: str
    val_expression: str
    val_default: float = 1.0

    def __post_init__(self):
        """Perform certain attribute quality assertions"""

        with suppress(NameError):
            eval(self.val_expression)


@dataclass
class Observable:

    name: str
    val_expression: Union[str, Callable]
    val_default: float = None
    is_required: bool = False

    def __post_init__(self):
        """Perform certain attribute quality assertions"""

        if isinstance(self.val_expression, str) is False:
            return

        with suppress(NameError):
            eval(self.val_expression)



from contextlib import suppress
from dataclasses import dataclass
from typing import Tuple


@dataclass
class AnalysisParameter:

    name: str
    lha_block: str
    lha_id: int
    max_power: int
    val_range: Tuple[float, float]
    transform: str

    def __str__(self) -> str:
        """
        Formats the analysis parameter in a nice way

        Returns
        -------
            Formatted parameter string
        """

        return (
            f"{self.name} "
            f"("
            f"LHA: {self.lha_block} {self.lha_id}, "
            f"Power: {self.max_power}, "
            f"Range: {self.val_range}"
            f")"
        )

    def __post_init__(self):
        """Perform certain attribute quality assertions"""

        with suppress(NameError):
            eval(self.transform)


@dataclass
class NuisanceParameter:

    name: str
    systematic: str
    benchmark_pos: str = None
    benchmark_neg: str = None

    def __str__(self) -> str:
        """
        Formats the nuisance parameter in a nice way

        Returns
        -------
            Formatted parameter string
        """

        return (
            f"{self.name} "
            f"("
            f"Systematic: {self.systematic}, "
            f"Benchmarks: {self.benchmark_pos} | {self.benchmark_neg}"
            f")"
        )



import logging

from collections import OrderedDict
from typing import Dict

from madminer.models import AnalysisParameter
from madminer.models import Benchmark
from madminer.models import Systematic
from madminer.models import SystematicScale
from madminer.models import SystematicType


logger = logging.getLogger(__name__)


def export_param_card(
    benchmark: Benchmark,
    parameters: Dict[str, AnalysisParameter],
    param_card_template_file: str,
    mg_process_directory: str,
    param_card_filename: str = None,
):
    # Open parameter card template
    with open(param_card_template_file) as file:
        param_card = file.read()
    lines = param_card.splitlines()

    # Replace parameter values
    for param_name, param_value in benchmark.values.items():
        param_lha_block = parameters[param_name].lha_block
        param_lha_id = parameters[param_name].lha_id
        param_transform = parameters[param_name].transform

        # Transform parameters if needed
        if param_transform is not None:
            variables = {"theta": param_value}
            param_value = eval(param_transform, variables)

        param_value = float(param_value)

        # Find entry
        current_block = None
        changed_line = False
        for i, line in enumerate(lines):

            # Remove comment
            try:
                line = line.split("#")[0]
            except:
                pass

            elements = line.split()

            # See if block begin
            if len(elements) == 2 and elements[0].lower() == "block":
                current_block = elements[1].lower()

            elif len(elements) == 2 and param_lha_block.lower() == current_block:
                try:
                    lha_id = int(elements[0])
                except ValueError:
                    continue

                if lha_id == param_lha_id:
                    lines[i] = f"    {param_lha_id}    {param_value}    # MadMiner"
                    changed_line = True
                    break

            elif len(elements) == 3 and elements[0].lower() == param_lha_block.lower():
                try:
                    lha_id = int(elements[1])
                except ValueError:
                    continue

                current_block = None
                if lha_id == param_lha_id:
                    lines[i] = f"{param_lha_block}    {param_lha_id}    {param_value}    # MadMiner"
                    changed_line = True
                    break

        if not changed_line:
            raise ValueError(f"Could not find LHA ID {param_lha_id} in param_card template!")

        param_card = "\n".join(lines)

    # Output filename
    if param_card_filename is None:
        param_card_filename = f"{mg_process_directory}/Cards/param_card.dat"

    # Save param_card.dat
    with open(param_card_filename, "w") as file:
        file.write(param_card)


def export_reweight_card(
    sample_benchmark: Benchmark,
    benchmarks: Dict[str, Benchmark],
    parameters: Dict[str, AnalysisParameter],
    mg_process_directory: str,
    reweight_card_filename: str = None,
):
    # Global setup
    lines = [
        "# Reweight card generated by MadMiner",
        "",
        "# Global setup",
        "change output default",
        "change helicity False",
    ]

    for benchmark_name, benchmark in benchmarks.items():
        if benchmark_name == sample_benchmark:
            continue

        lines.append("")
        lines.append("# MadMiner benchmark " + benchmark_name)
        lines.append("launch --rwgt_name=" + benchmark_name)

        for param_name, param_value in benchmark.values.items():
            param_lha_block = parameters[param_name].lha_block
            param_lha_id = parameters[param_name].lha_id
            param_transform = parameters[param_name].transform

            # Transform parameters if needed
            if param_transform is not None:
                variables = {"theta": param_value}
                param_value = eval(param_transform, variables)

            lines.append(f"  set {param_lha_block} {param_lha_id} {param_value}")

        lines.append("")

    reweight_card = "\n".join(lines)

    # Output filename
    if reweight_card_filename is None:
        reweight_card_filename = f"{mg_process_directory}/Cards/reweight_card.dat"

    # Save param_card.dat
    with open(reweight_card_filename, "w") as file:
        file.write(reweight_card)


def export_run_card(
    template_filename: str,
    run_card_filename: str,
    systematics: Dict[str, Systematic] = None,
    order: str = "LO",
):
    # Open parameter card template
    with open(template_filename) as file:
        run_card_template = file.read()

    run_card_lines = run_card_template.split("\n")

    # Do we actually have to run MadGraph's systematics feature?
    run_systematics = False
    for syst in systematics.values():
        if syst.type in {SystematicType.PDF, SystematicType.SCALE}:
            run_systematics = True
            break

    # Lines to be removed
    entries_to_comment_out = [
        "use_syst",
        "systematics_program",
        "systematics_argument",
        "systematics_arguments",
        "sys_scalefact",
        "sys_alpsfact",
        "sys_matchscale",
        "sys_pdf",
    ]

    # Changes to be made
    settings = OrderedDict()
    settings["use_syst"] = "False"
    if run_systematics:
        settings["use_syst"] = "True"
        settings["systematics_program"] = "systematics"
        settings["systematics_arguments"] = create_systematics_arguments(systematics)

    # Remove old entries
    for i, line in enumerate(run_card_lines):
        line_content = line
        # Remove comments
        try:
            line_content = line_content.split("#")[0]
        except:
            pass
        try:
            line_content = line_content.split("!")[0]
        except:
            pass

        # Split at last equal sign
        elements = line_content.split("=")
        if len(elements) < 2:
            continue
        line_key = elements[-1].strip()

        if line_key in entries_to_comment_out:
            run_card_lines[i] = f"# {line} # Commented out by MadMiner"
            continue

    # Add new entries - systematics
    if order == "LO":
        run_card_lines.append("")
        run_card_lines.append("#*********************************************************************")
        run_card_lines.append("# MadMiner systematics setup                                         *")
        run_card_lines.append("#*********************************************************************")
        for key, value in settings.items():
            run_card_lines.append(f"{value} = {key}")
        run_card_lines.append("")

    # Write new run card
    new_run_card = "\n".join(run_card_lines)
    with open(run_card_filename, "w") as file:
        file.write(new_run_card)


def create_systematics_arguments(systematics: Dict[str, Systematic]):
    """ Put together systematics_arguments string for MadGraph run card """

    if systematics is None:
        return ""

    systematics_arguments = []

    mur_done = False
    muf_done = False
    pdf_done = False

    for systematic in systematics.values():

        if systematic.type == SystematicType.SCALE and systematic.scale == SystematicScale.MU:
            if mur_done or muf_done:
                raise ValueError("Multiple nuisance parameter for scale variation!")
            systematics_arguments.append(f"'--mur={systematic.value}'")
            systematics_arguments.append(f"'--muf={systematic.value}'")
            systematics_arguments.append(f"'--together=mur,muf'")
            systematics_arguments.append(f"'--dyn=-1'")
            mur_done = True
            muf_done = True

        elif systematic.type == SystematicType.SCALE and systematic.scale == SystematicScale.MUR:
            if mur_done:
                raise ValueError("Multiple nuisance parameter for mur variation!")
            systematics_arguments.append(f"'--mur={systematic.value}'")
            systematics_arguments.append(f"'--dyn=-1'")
            mur_done = True

        elif systematic.type == SystematicType.SCALE and systematic.scale == SystematicScale.MUF:
            if muf_done:
                raise ValueError("Multiple nuisance parameter for muf variation!")
            systematics_arguments.append(f"'--muf={systematic.value}'")
            systematics_arguments.append(f"'--dyn=-1'")
            muf_done = True

        elif systematic.type == SystematicType.PDF:
            if pdf_done:
                raise ValueError("Multiple nuisance parameter for PDF variation!")
            systematics_arguments.append(f"'--pdf={systematic.value}'")
            pdf_done = True

    if len(systematics_arguments) > 0:
        return f"[{', '.join(systematics_arguments)}]"

    return ""



import logging
import numpy as np

from collections import OrderedDict

from .base import ConditionalEstimator
from ..utils.ml.eval import evaluate_flow_model
from ..utils.ml.models.maf import ConditionalMaskedAutoregressiveFlow
from ..utils.ml.models.maf_mog import ConditionalMixtureMaskedAutoregressiveFlow
from ..utils.ml.trainer import FlowTrainer
from ..utils.ml.utils import get_optimizer, get_loss
from ..utils.various import load_and_check, shuffle, restrict_samplesize


logger = logging.getLogger(__name__)


class LikelihoodEstimator(ConditionalEstimator):
    """A neural estimator of the density or likelihood evaluated at a reference hypothesis as a function
     of the observation x.

    Parameters
    ----------
    features : list of int or None, optional
        Indices of observables (features) that are used as input to the neural networks. If None, all observables
        are used. Default value: None.

    n_components : int, optional
        The number of Gaussian base components in a MADE MoG. If 1, a plain MADE is used.
        Default value: 1.

    n_mades : int, optional
        The number of MADE layers. Default value: 3.

    n_hidden : tuple of int, optional
        Units in each hidden layer in the neural networks. If method is 'nde' or 'scandal', this refers to the
        setup of each individual MADE layer. Default value: (100,).

    activation : {'tanh', 'sigmoid', 'relu'}, optional
        Activation function. Default value: 'tanh'.

    batch_norm : None or float, optional
        If not None, batch normalization is used, where this value sets the alpha parameter in the calculation
        of the running average of the mean and variance. Default value: None.
    """

    def __init__(self, features=None, n_components=1, n_mades=5, n_hidden=(100,), activation="tanh", batch_norm=None):
        super().__init__(features, n_hidden, activation, dropout_prob=0.0)

        self.n_components = n_components
        self.n_mades = n_mades
        self.batch_norm = batch_norm

    def train(
        self,
        method,
        x,
        theta,
        t_xz=None,
        x_val=None,
        theta_val=None,
        t_xz_val=None,
        alpha=1.0,
        optimizer="amsgrad",
        n_epochs=50,
        batch_size=128,
        initial_lr=0.001,
        final_lr=0.0001,
        nesterov_momentum=None,
        validation_split=0.25,
        early_stopping=True,
        scale_inputs=True,
        shuffle_labels=False,
        limit_samplesize=None,
        memmap=False,
        verbose="some",
        scale_parameters=True,
        n_workers=8,
        clip_gradient=None,
        early_stopping_patience=None,
    ):
        """
        Trains the network.

        Parameters
        ----------
        method : str
            The inference method used for training. Allowed values are 'nde' and 'scandal'.

        x : ndarray or str
            Observations, or filename of a pickled numpy array.

        theta : ndarray or str
            Numerator parameter point, or filename of a pickled numpy array.

        t_xz : ndarray or str or None, optional
            Joint scores at theta, or filename of a pickled numpy array. Default value: None.

        x_val : ndarray or str or None, optional
            Validation observations, or filename of a pickled numpy array. If None
            and validation_split > 0, validation data will be randomly selected from the training data.
            Default value: None.

        theta_val : ndarray or str or None, optional
            Validation numerator parameter points, or filename of a pickled numpy array. If None
            and validation_split > 0, validation data will be randomly selected from the training data.
            Default value: None.

        t_xz_val : ndarray or str or None, optional
            Validation joint scores at theta, or filename of a pickled numpy array. If None
            and validation_split > 0, validation data will be randomly selected from the training data.
            Default value: None.

        alpha : float, optional
            Hyperparameter weighting the score error in the loss function of the 'alices', 'rascal', and 'cascal'
            methods. Default value: 1.

        optimizer : {"adam", "amsgrad", "sgd"}, optional
            Optimization algorithm. Default value: "amsgrad".

        n_epochs : int, optional
            Number of epochs. Default value: 50.

        batch_size : int, optional
            Batch size. Default value: 128.

        initial_lr : float, optional
            Learning rate during the first epoch, after which it exponentially decays to final_lr. Default value:
            0.001.

        final_lr : float, optional
            Learning rate during the last epoch. Default value: 0.0001.

        nesterov_momentum : float or None, optional
            If trainer is "sgd", sets the Nesterov momentum. Default value: None.

        validation_split : float or None, optional
            Fraction of samples used  for validation and early stopping (if early_stopping is True). If None, the entire
            sample is used for training and early stopping is deactivated. Default value: 0.25.

        early_stopping : bool, optional
            Activates early stopping based on the validation loss (only if validation_split is not None). Default value:
            True.

        scale_inputs : bool, optional
            Scale the observables to zero mean and unit variance. Default value: True.

        shuffle_labels : bool, optional
            If True, the labels (`y`, `r_xz`, `t_xz`) are shuffled, while the observations (`x`) remain in their
            normal order. This serves as a closure test, in particular as cross-check against overfitting: an estimator
            trained with shuffle_labels=True should predict to likelihood ratios around 1 and scores around 0.

        limit_samplesize : int or None, optional
            If not None, only this number of samples (events) is used to train the estimator. Default value: None.

        memmap : bool, optional.
            If True, training files larger than 1 GB will not be loaded into memory at once. Default value: False.

        verbose : {"all", "many", "some", "few", "none}, optional
            Determines verbosity of training. Default value: "some".

        scale_parameters : bool, optional
            Whether parameters are rescaled to mean zero and unit variance before going into the neural network.
            Default value: True.

        Returns
        -------
        result: ndarray
            Training and validation losses from FlowTrainer.train
        """

        logger.info("Starting training")
        logger.info("  Method:                 %s", method)
        if method == "scandal":
            logger.info("  alpha:                  %s", alpha)
        logger.info("  Batch size:             %s", batch_size)
        logger.info("  Optimizer:              %s", optimizer)
        logger.info("  Epochs:                 %s", n_epochs)
        logger.info("  Learning rate:          %s initially, decaying to %s", initial_lr, final_lr)
        if optimizer == "sgd":
            logger.info("  Nesterov momentum:      %s", nesterov_momentum)
        logger.info("  Validation split:       %s", validation_split)
        logger.info("  Early stopping:         %s", early_stopping)
        logger.info("  Scale inputs:           %s", scale_inputs)
        logger.info("  Shuffle labels          %s", shuffle_labels)
        if limit_samplesize is None:
            logger.info("  Samples:                all")
        else:
            logger.info("  Samples:                %s", limit_samplesize)

        # Load training data
        logger.info("Loading training data")
        memmap_threshold = 1.0 if memmap else None
        theta = load_and_check(theta, memmap_files_larger_than_gb=memmap_threshold)
        x = load_and_check(x, memmap_files_larger_than_gb=memmap_threshold)
        t_xz = load_and_check(t_xz, memmap_files_larger_than_gb=memmap_threshold)

        self._check_required_data(method, t_xz)

        # Infer dimensions of problem
        n_samples = x.shape[0]
        n_observables = x.shape[1]
        n_parameters = theta.shape[1]
        logger.info("Found %s samples with %s parameters and %s observables", n_samples, n_parameters, n_observables)

        # Limit sample size
        if limit_samplesize is not None and limit_samplesize < n_samples:
            logger.info("Only using %s of %s training samples", limit_samplesize, n_samples)
            x, theta, t_xz = restrict_samplesize(limit_samplesize, x, theta, t_xz)

        # Validation data
        external_validation = x_val is not None and theta_val is not None
        if external_validation:
            theta_val = load_and_check(theta_val, memmap_files_larger_than_gb=memmap_threshold)
            x_val = load_and_check(x_val, memmap_files_larger_than_gb=memmap_threshold)
            t_xz_val = load_and_check(t_xz_val, memmap_files_larger_than_gb=memmap_threshold)

            logger.info("Found %s separate validation samples", x_val.shape[0])

            assert x_val.shape[1] == n_observables
            assert theta_val.shape[1] == n_parameters
            if t_xz is not None:
                assert t_xz_val is not None, "When providing t_xz and sep. validation data, also provide t_xz_val"

        # Scale features
        if scale_inputs:
            self.initialize_input_transform(x, overwrite=False)
            x = self._transform_inputs(x)
            if external_validation:
                x_val = self._transform_inputs(x_val)
        else:
            self.initialize_input_transform(x, False, overwrite=False)

        # Scale parameters
        if scale_parameters:
            logger.info("Rescaling parameters")
            self.initialize_parameter_transform(theta)
            theta = self._transform_parameters(theta)
            t_xz = self._transform_score(t_xz, inverse=False)
            if external_validation:
                t_xz_val = self._transform_score(t_xz_val, inverse=False)
        else:
            self.initialize_parameter_transform(theta, False)

        # Shuffle labels
        if shuffle_labels:
            logger.info("Shuffling labels")
            t_xz = shuffle(t_xz)

        # Features
        if self.features is not None:
            x = x[:, self.features]
            logger.info("Only using %s of %s observables", x.shape[1], n_observables)
            n_observables = x.shape[1]
            if external_validation:
                x_val = x_val[:, self.features]

        # Check consistency of input with model
        if self.n_observables is None:
            self.n_observables = n_observables
        if self.n_parameters is None:
            self.n_parameters = n_parameters

        if n_parameters != self.n_parameters:
            raise RuntimeError(f"Number of parameters does not match: {n_parameters} vs {self.n_parameters}")
        if n_observables != self.n_observables:
            raise RuntimeError(f"Number of observables does not match: {n_observables} vs {self.n_observables}")

        # Data
        data = self._package_training_data(method, x, theta, t_xz)
        if external_validation:
            data_val = self._package_training_data(method, x_val, theta_val, t_xz_val)
        else:
            data_val = None

        # Create model
        if self.model is None:
            self._create_model()

        # Losses
        loss_functions, loss_labels, loss_weights = get_loss(method, alpha)

        # Optimizer
        opt, opt_kwargs = get_optimizer(optimizer, nesterov_momentum)

        # Train model
        logger.info("Training model")
        trainer = FlowTrainer(self.model, n_workers=n_workers)
        result = trainer.train(
            data=data,
            data_val=data_val,
            loss_functions=loss_functions,
            loss_weights=loss_weights,
            loss_labels=loss_labels,
            epochs=n_epochs,
            batch_size=batch_size,
            optimizer=opt,
            optimizer_kwargs=opt_kwargs,
            initial_lr=initial_lr,
            final_lr=final_lr,
            validation_split=validation_split,
            early_stopping=early_stopping,
            verbose=verbose,
            clip_gradient=clip_gradient,
            early_stopping_patience=early_stopping_patience,
        )
        return result

    def evaluate_log_likelihood(self, x, theta, test_all_combinations=True, evaluate_score=False):
        """
        Evaluates the log likelihood as a function of the observation x and the parameter point theta.

        Parameters
        ----------
        x : ndarray or str
            Sample of observations, or path to numpy file with observations.

        theta : ndarray or str
            Parameter points, or path to numpy file with parameter points.

        test_all_combinations : bool, optional
            If method is not 'sally' and not 'sallino': If False, the number of samples in the observable and theta
            files has to match, and the likelihood ratio is evaluated only for the combinations
            `r(x_i | theta0_i, theta1_i)`. If True, `r(x_i | theta0_j, theta1_j)` for all pairwise combinations `i, j`
            are evaluated. Default value: True.

        evaluate_score : bool, optional
            If method is not 'sally' and not 'sallino', this sets whether in addition to the likelihood ratio the score
            is evaluated. Default value: False.

        Returns
        -------

        log_likelihood : ndarray
            The estimated log likelihood. If test_all_combinations is True, the result has shape `(n_thetas, n_x)`.
            Otherwise, it has shape `(n_samples,)`.

        score : ndarray or None
            None if
            evaluate_score is False. Otherwise the derived estimated score at `theta`. If test_all_combinations is
            True, the result has shape `(n_thetas, n_x, n_parameters)`. Otherwise, it has shape
            `(n_samples, n_parameters)`.
        """

        if self.model is None:
            raise ValueError("No model -- train or load model before evaluating it!")

        # Load training data
        if isinstance(x, str):
            logger.debug("Loading evaluation data")
        theta = load_and_check(theta)
        x = load_and_check(x)

        # Scale observables
        x = self._transform_inputs(x)

        # Restrict features
        if self.features is not None:
            x = x[:, self.features]

        # Evaluation for all other methods
        all_log_p_hat = []
        all_t_hat = []

        if test_all_combinations:
            logger.debug("Starting ratio evaluation for %s x-theta combinations", len(theta) * len(x))

            for i, this_theta in enumerate(theta, start=1):
                logger.debug("Starting log likelihood evaluation for thetas %s / %s: %s", i, len(theta), this_theta)

                log_p_hat, t_hat = evaluate_flow_model(
                    model=self.model,
                    thetas=[this_theta],
                    xs=x,
                    evaluate_score=evaluate_score,
                )

                all_log_p_hat.append(log_p_hat)
                all_t_hat.append(t_hat)

            all_log_p_hat = np.array(all_log_p_hat)
            all_t_hat = np.array(all_t_hat)

        else:
            logger.debug("Starting log likelihood evaluation")

            all_log_p_hat, all_t_hat = evaluate_flow_model(
                model=self.model, thetas=theta, xs=x, evaluate_score=evaluate_score
            )

        logger.debug("Evaluation done")
        return all_log_p_hat, all_t_hat

    def evaluate_log_likelihood_ratio(self, x, theta0, theta1, test_all_combinations, evaluate_score=False):
        """
        Evaluates the log likelihood ratio as a function of the observation x, the numerator parameter point theta0,
        and the denominator parameter point theta1.

        Parameters
        ----------
        x : ndarray or str
            Sample of observations, or path to numpy file with observations.

        theta0 : ndarray or str
            Numerator parameters, or path to numpy file.

        theta1 : ndarray or str
            Denominator parameters, or path to numpy file.

        test_all_combinations : bool, optional
            If method is not 'sally' and not 'sallino': If False, the number of samples in the observable and theta
            files has to match, and the likelihood ratio is evaluated only for the combinations
            `r(x_i | theta0_i, theta1_i)`. If True, `r(x_i | theta0_j, theta1_j)` for all pairwise combinations `i, j`
            are evaluated. Default value: True.

        evaluate_score : bool, optional
            If method is not 'sally' and not 'sallino', this sets whether in addition to the likelihood ratio the score
            is evaluated. Default value: False.

        Returns
        -------

        log_likelihood : ndarray
            The estimated log likelihood. If test_all_combinations is True, the result has shape `(n_thetas, n_x)`.
            Otherwise, it has shape `(n_samples,)`.

        score : ndarray or None
            None if
            evaluate_score is False. Otherwise the derived estimated score at `theta`. If test_all_combinations is
            True, the result has shape `(n_thetas, n_x, n_parameters)`. Otherwise, it has shape
            `(n_samples, n_parameters)`.
        """

        if self.model is None:
            raise ValueError("No model -- train or load model before evaluating it!")

        # Load training data
        logger.debug("Loading evaluation data")
        x = load_and_check(x)
        theta0 = load_and_check(theta0)
        theta1 = load_and_check(theta1)

        # Scale observables
        x = self._transform_inputs(x)

        # Restrict features
        if self.features is not None:
            x = x[:, self.features]

        # Balance thetas
        if len(theta1) > len(theta0):
            theta0 = [theta0[i % len(theta0)] for i in range(len(theta1))]
        elif len(theta1) < len(theta0):
            theta1 = [theta1[i % len(theta1)] for i in range(len(theta0))]

        log_p_hat0, t_hat0 = self.evaluate_log_likelihood(
            x, theta0, test_all_combinations=test_all_combinations, evaluate_score=evaluate_score
        )
        log_p_hat1, t_hat1 = self.evaluate_log_likelihood(
            x, theta1, test_all_combinations=test_all_combinations, evaluate_score=evaluate_score
        )
        log_r_hat = log_p_hat0 - log_p_hat1

        return log_r_hat, t_hat0, t_hat1

    def evaluate_score(self, *args, **kwargs):
        raise NotImplementedError("Please use evaluate_log_likelihood(evaluate_score=True).")

    def calculate_fisher_information(self, *args, **kwargs):
        raise NotImplementedError(
            "Please use evaluate_log_likelihood_ratio(evaluate_score=True) and calculate the "
            "Fisher information manually."
        )

    def evaluate(self, *args, **kwargs):
        return self.evaluate_log_likelihood(*args, **kwargs)

    def _create_model(self):
        if self.n_components > 1:
            self.model = ConditionalMixtureMaskedAutoregressiveFlow(
                n_conditionals=self.n_parameters,
                n_inputs=self.n_observables,
                n_components=self.n_components,
                n_hiddens=self.n_hidden,
                n_mades=self.n_mades,
                activation=self.activation,
                batch_norm=self.batch_norm is not None,
                alpha=self.batch_norm,
            )
        else:
            self.model = ConditionalMaskedAutoregressiveFlow(
                n_conditionals=self.n_parameters,
                n_inputs=self.n_observables,
                n_hiddens=self.n_hidden,
                n_mades=self.n_mades,
                activation=self.activation,
                batch_norm=self.batch_norm is not None,
                alpha=self.batch_norm,
            )

    @staticmethod
    def _check_required_data(method, t_xz):
        if method == ["scandal"] and t_xz is None:
            raise RuntimeError(f"Method {method} requires joint score information")

    @staticmethod
    def _package_training_data(method, x, theta, t_xz):
        data = OrderedDict()
        data["x"] = x
        data["theta"] = theta
        if method in ["scandal"]:
            data["t_xz"] = t_xz
        return data

    def _wrap_settings(self):
        settings = super()._wrap_settings()
        settings["estimator_type"] = "likelihood"
        settings["n_components"] = self.n_components
        settings["batch_norm"] = self.batch_norm
        settings["n_mades"] = self.n_mades
        return settings

    def _unwrap_settings(self, settings):
        super()._unwrap_settings(settings)

        estimator_type = str(settings["estimator_type"])
        if estimator_type != "likelihood":
            raise RuntimeError(f"Saved model is an incompatible estimator type {estimator_type}.")

        self.n_components = int(settings["n_components"])
        self.n_mades = int(settings["n_mades"])
        self.batch_norm = settings["batch_norm"]
        if self.batch_norm == "None":
            self.batch_norm = None
        if self.batch_norm is not None:
            self.batch_norm = float(self.batch_norm)



import logging
import numpy as np
import torch
import torch.nn as nn

from madminer.utils.ml.utils import get_activation_function, check_for_nan, check_for_nonpos, NanException
from torch.autograd import grad

logger = logging.getLogger(__name__)


class DenseSingleParameterizedRatioModel(nn.Module):
    """Module that implements agnostic parameterized likelihood estimators such as RASCAL or ALICES. Only the
    numerator of the ratio is parameterized."""

    def __init__(self, n_observables, n_parameters, n_hidden, activation="tanh", dropout_prob=0.0):
        super().__init__()

        # Save input
        self.n_hidden = n_hidden
        self.activation = get_activation_function(activation)
        self.dropout_prob = dropout_prob

        # Build network
        self.layers = nn.ModuleList()
        n_last = n_observables + n_parameters

        # Hidden layers
        for n_hidden_units in n_hidden:
            if self.dropout_prob > 1.0e-9:
                self.layers.append(nn.Dropout(self.dropout_prob))
            self.layers.append(nn.Linear(n_last, n_hidden_units))
            n_last = n_hidden_units

        # Log r layer
        if self.dropout_prob > 1.0e-9:
            self.layers.append(nn.Dropout(self.dropout_prob))
        self.append = self.layers.append(nn.Linear(n_last, 1))

    def forward(self, theta, x, track_score=True, return_grad_x=False, create_gradient_graph=True):
        """ Calculates estimated log likelihood ratio and the derived score. """

        # Track gradient wrt theta
        if track_score and not theta.requires_grad:
            theta.requires_grad = True

        # Track gradient wrt x
        if return_grad_x and not x.requires_grad:
            x.requires_grad = True

        # log r estimator
        log_r_hat = torch.cat((theta, x), 1)

        for i, layer in enumerate(self.layers):
            if i > 0:
                log_r_hat = self.activation(log_r_hat)
            log_r_hat = layer(log_r_hat)

        # Bayes-optimal s
        s_hat = 1.0 / (1.0 + torch.exp(log_r_hat))

        # Score t
        if track_score:
            (t_hat,) = grad(
                log_r_hat,
                theta,
                grad_outputs=torch.ones_like(log_r_hat.data),
                # grad_outputs=log_r_hat.data.new(log_r_hat.shape).fill_(1),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )
        else:
            t_hat = None

        # Calculate gradient wrt x
        if return_grad_x:
            (x_gradient,) = grad(
                log_r_hat,
                x,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )

            return s_hat, log_r_hat, t_hat, x_gradient

        return s_hat, log_r_hat, t_hat

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)

        for i, layer in enumerate(self.layers):
            self.layers[i] = layer.to(*args, **kwargs)

        return self


class DenseDoublyParameterizedRatioModel(nn.Module):
    """Module that implements agnostic parameterized likelihood estimators such as RASCAL or ALICES. Both
    numerator and denominator of the ratio are parameterized."""

    def __init__(self, n_observables, n_parameters, n_hidden, activation="tanh", dropout_prob=0.0):
        super().__init__()

        # Save input
        self.n_hidden = n_hidden
        self.activation = get_activation_function(activation)
        self.dropout_prob = dropout_prob

        # Build network
        self.layers = nn.ModuleList()
        n_last = n_observables + 2 * n_parameters

        # Hidden layers
        for n_hidden_units in n_hidden:
            if self.dropout_prob > 1.0e-9:
                self.layers.append(nn.Dropout(self.dropout_prob))
            self.layers.append(nn.Linear(n_last, n_hidden_units))
            n_last = n_hidden_units

        # Log r layer
        if self.dropout_prob > 1.0e-9:
            self.layers.append(nn.Dropout(self.dropout_prob))
        self.layers.append(nn.Linear(n_last, 1))

    def forward(self, theta0, theta1, x, track_score=True, return_grad_x=False, create_gradient_graph=True):
        """ Calculates estimated log likelihood ratio and the derived score. """

        # Track gradient wrt thetas
        if track_score and not theta0.requires_grad:
            theta0.requires_grad = True
        if track_score and not theta1.requires_grad:
            theta1.requires_grad = True

        # Track gradient wrt x
        if return_grad_x and not x.requires_grad:
            x.requires_grad = True

        # f(x | theta0, theta1)
        f_th0_th1 = torch.cat((theta0, theta1, x), 1)

        for i, layer in enumerate(self.layers):
            if i > 0:
                f_th0_th1 = self.activation(f_th0_th1)
            f_th0_th1 = layer(f_th0_th1)

        # f(x | theta1, theta0)
        f_th1_th0 = torch.cat((theta1, theta0, x), 1)

        for i, layer in enumerate(self.layers):
            if i > 0:
                f_th1_th0 = self.activation(f_th1_th0)
            f_th1_th0 = layer(f_th1_th0)

        # Antisymmetric combination
        log_r_hat = f_th0_th1 - f_th1_th0

        # Bayes-optimal s
        s_hat = 1.0 / (1.0 + torch.exp(log_r_hat))

        # Score t
        if track_score:
            (t_hat0,) = grad(
                log_r_hat,
                theta0,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )
            (t_hat1,) = grad(
                log_r_hat,
                theta1,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )
            # NOTE: this is a factor of 4 slower than the simple parameterized version (2 gradients * 2 times
            #       slower calculation each)
        else:
            t_hat0 = None
            t_hat1 = None

        # Calculate gradient wrt x
        if return_grad_x:
            (x_gradient,) = grad(
                log_r_hat,
                x,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )

            return s_hat, log_r_hat, t_hat0, t_hat1, x_gradient

        return s_hat, log_r_hat, t_hat0, t_hat1

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)

        for i, layer in enumerate(self.layers):
            self.layers[i] = layer.to(*args, **kwargs)

        return self


class DenseComponentRatioModel(nn.Module):
    """Module that implements agnostic parameterized likelihood estimators such as RASCAL or ALICES. Only the
    numerator of the ratio is parameterized."""

    def __init__(self, n_observables, n_hidden, activation="tanh", dropout_prob=0.0):
        super().__init__()

        # Save input
        self.n_hidden = n_hidden
        self.activation = get_activation_function(activation)
        self.dropout_prob = dropout_prob

        # Build network
        self.layers = nn.ModuleList()
        n_last = n_observables

        # Hidden layers
        for n_hidden_units in n_hidden:
            if self.dropout_prob > 1.0e-9:
                self.layers.append(nn.Dropout(self.dropout_prob))
            self.layers.append(nn.Linear(n_last, n_hidden_units))
            n_last = n_hidden_units

        # Log r layer
        if self.dropout_prob > 1.0e-9:
            self.layers.append(nn.Dropout(self.dropout_prob))
        self.layers.append(nn.Linear(n_last, 1))

    def forward(self, x):

        # log r estimator
        log_r_hat = x

        for i, layer in enumerate(self.layers):
            if i > 0:
                log_r_hat = self.activation(log_r_hat)
            log_r_hat = layer(log_r_hat)

        return log_r_hat

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)

        for i, layer in enumerate(self.layers):
            self.layers[i] = layer.to(*args, **kwargs)

        return self


class DenseMorphingAwareRatioModel(nn.Module):

    def __init__(
        self,
        components,
        morphing_matrix,
        n_observables,
        n_parameters,
        n_hidden,
        activation="tanh",
        dropout_prob=0.0,
        clamp_component_ratios=5.0,
    ):

        super().__init__()

        # Save input
        self.n_hidden = n_hidden
        self.activation = get_activation_function(activation)
        self.dropout_prob = dropout_prob
        self.n_components, self.n_parameters = components.shape
        self.clamp_component_ratios = clamp_component_ratios

        # Morphing setup
        self.register_buffer("components", torch.tensor(components))
        self.register_buffer("morphing_matrix", torch.tensor(morphing_matrix))

        logger.debug("Loaded morphing matrix into PyTorch model:\n %s", morphing_matrix)

        # Build networks for all components
        self.dsigma_component_estimators = nn.ModuleList()
        for _ in range(self.n_components):
            self.dsigma_component_estimators.append(
                DenseComponentRatioModel(n_observables, n_hidden, activation, dropout_prob)
            )

        self.log_sigma_ratio_components = torch.nn.Parameter(torch.zeros((self.n_components, 1)))  # (n_components, 1)

    def forward(self, theta, x, track_score=True, return_grad_x=False, create_gradient_graph=True):
        """ Calculates estimated log likelihood ratio and the derived score. """

        # Track gradient wrt theta
        if track_score and not theta.requires_grad:
            theta.requires_grad = True

        # Track gradient wrt x
        if return_grad_x and not x.requires_grad:
            x.requires_grad = True

        # Calculate individual components dsigma_c(x) / dsigma1
        dsigma_ratio_components = [component(x).unsqueeze(1) for component in self.dsigma_component_estimators]
        dsigma_ratio_components = torch.cat(dsigma_ratio_components, 1)  # (batchsize, n_components, 1)
        dsigma_ratio_components = torch.exp(
            torch.clamp(dsigma_ratio_components, -self.clamp_component_ratios, self.clamp_component_ratios)
        )

        # Denominator (for changes in total xsecs with theta)
        sigma_ratio_components = torch.exp(
            torch.clamp(self.log_sigma_ratio_components, -self.clamp_component_ratios, self.clamp_component_ratios)
        )

        # Calculate morphing weights
        component_weights = []
        for c in range(self.n_components):
            component_weight = 1.0
            for p in range(self.n_parameters):
                component_weight = component_weight * (theta[:, p] ** self.components[c, p])
            component_weights.append(component_weight.unsqueeze(1))
        component_weights = torch.cat(component_weights, dim=1)  # (batchsize, n_components)

        # Put together
        weights = torch.einsum("cn,bc->bn", [self.morphing_matrix, component_weights])  # (batchsize, n_benchmarks)
        numerator = torch.einsum("bn,bno->bo", [weights, dsigma_ratio_components])
        denominator = torch.einsum("bn,no->bo", [weights, sigma_ratio_components])

        numerator = torch.clamp(numerator, np.exp(-self.clamp_component_ratios), np.exp(self.clamp_component_ratios))
        denominator = torch.clamp(
            denominator, np.exp(-self.clamp_component_ratios), np.exp(self.clamp_component_ratios)
        )
        r_hat = numerator / denominator

        # # Debugging
        # try:
        #     check_for_nonpos("Morphing-aware model: numerator", numerator)
        #     check_for_nonpos("Morphing-aware model: denominator", denominator)
        #     check_for_nonpos("Morphing-aware model: ratio", r_hat)
        # except NanException:
        #     logger.error("Inconsistent inputs in the morphing-aware model.")
        #
        #     filter = torch.where((r_hat <= 0.).any(axis = 1))
        #
        #     logger.error(  "x = %s", x[filter])
        #     logger.error(  "theta = %s", theta[filter])
        #     logger.error(  "morphing_matrix = %s", self.morphing_matrix)
        #     logger.error(  "dsigma_ratio_components = %s", dsigma_ratio_components[filter])
        #     logger.error(  "sigma_ratio_components = %s", sigma_ratio_components)
        #     logger.error(  "component_weights = %s", component_weights[filter])
        #     logger.error(  "numerator = %s", numerator[filter])
        #     logger.error(  "denominator = %s", denominator[filter])
        #     logger.error(  "ratio = %s", r_hat[filter])
        #
        #     logger.error("Let's go into this calculation a bit more:")
        #     # morphing_matrix:  (n_benchmarks, n_components)
        #     weights = torch.einsum("cn,bc->bn", [self.morphing_matrix, component_weights])  # (batchsize, n_benchmarks)
        #     logger.error("Thetas -> weights:")
        #     for i in range(weights.size(0)):
        #         logger.error("  %s -> %s", theta[i].detach().numpy(), weights[i].detach().numpy())
        #     logger.error("Weights: %s", weights.detach().numpy())
        #     logger.error("Component predictions for numerator: %s", sigma_ratio_components.detach().numpy())
        #     logger.error("Combined numerator: %s", numerator.detach().numpy())
        #
        #     raise

        # Bayes-optimal s
        r_hat = torch.clamp(r_hat, np.exp(-self.clamp_component_ratios), np.exp(self.clamp_component_ratios))
        log_r_hat = torch.log(torch.clamp(r_hat, 1.0e-6, 1.0e6))
        s_hat = 1.0 / (1.0 + r_hat)

        # Score t
        if track_score:
            (t_hat,) = grad(
                log_r_hat,
                theta,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )
        else:
            t_hat = None

        # Calculate gradient wrt x
        if return_grad_x:
            (x_gradient,) = grad(
                log_r_hat,
                x,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )

            return s_hat, log_r_hat, t_hat, x_gradient

        return s_hat, log_r_hat, t_hat


class DenseQuadraticMorphingAwareRatioModel(nn.Module):

    def __init__(self, n_observables, n_parameters, n_hidden, activation="tanh", dropout_prob=0.0):
        super().__init__()

        assert n_parameters == 1

        # Save input
        self.n_hidden = n_hidden
        self.activation = get_activation_function(activation)
        self.dropout_prob = dropout_prob

        # Build networks for all components
        self.dsigma_a = DenseComponentRatioModel(n_observables, n_hidden, activation, dropout_prob)
        self.dsigma_b = DenseComponentRatioModel(n_observables, n_hidden, activation, dropout_prob)
        self.sigma_a = torch.nn.Parameter(torch.ones((1,)))
        self.sigma_b = torch.nn.Parameter(torch.ones((1,)))

    def forward(self, theta, x, track_score=True, return_grad_x=False, create_gradient_graph=True):
        """ Calculates estimated log likelihood ratio and the derived score. """

        assert theta.size()[1] == 1

        # Track gradient wrt theta
        if track_score and not theta.requires_grad:
            theta.requires_grad = True

        # Track gradient wrt x
        if return_grad_x and not x.requires_grad:
            x.requires_grad = True

        # Calculate individual components
        dsigma_a = self.dsigma_a(x)  # (batch, 1)
        dsigma_b = self.dsigma_b(x)  # (batch, 1)

        # Put together
        dsigma_ratio = (1 + theta * dsigma_a) ** 2 + (theta * dsigma_b) ** 2
        sigma_ratio = (1 + theta * self.sigma_a) ** 2 + (theta * self.sigma_b) ** 2
        r_hat = dsigma_ratio / sigma_ratio

        # Bayes-optimal s
        log_r_hat = torch.log(r_hat + 1.0e-9)
        s_hat = 1.0 / (1.0 + r_hat)

        # Score t
        if track_score:
            (t_hat,) = grad(
                log_r_hat,
                theta,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )
        else:
            t_hat = None

        # Calculate gradient wrt x
        if return_grad_x:
            (x_gradient,) = grad(
                log_r_hat,
                x,
                grad_outputs=torch.ones_like(log_r_hat.data),
                only_inputs=True,
                create_graph=create_gradient_graph,
            )

            return s_hat, log_r_hat, t_hat, x_gradient

        return s_hat, log_r_hat, t_hat



import logging
import numpy as np
import os

from collections import OrderedDict
from typing import Callable
from typing import Dict
from typing import List

try:
    import xml.etree.cElementTree as ET
    use_celementtree = True
except ImportError:
    import xml.etree.ElementTree as ET
    use_celementtree = False

from particle import Particle
from madminer.models import Cut
from madminer.models import Efficiency
from madminer.models import Observable
from madminer.models import Systematic
from madminer.models import SystematicScale
from madminer.models import SystematicType
from madminer.utils.particle import MadMinerParticle
from madminer.utils.various import unzip_file, approx_equal, math_commands

logger = logging.getLogger(__name__)


def parse_lhe_file(
    filename,
    sampling_benchmark,
    observables: Dict[str, Observable],
    cuts: List[Cut] = None,
    efficiencies: List[Efficiency] = None,
    benchmark_names=None,
    is_background=False,
    energy_resolutions=None,
    pt_resolutions=None,
    eta_resolutions=None,
    phi_resolutions=None,
    k_factor=1.0,
    parse_events_as_xml=True,
    systematics_dict=None,
):
    """ Extracts observables and weights from a LHE file """

    logger.debug("Parsing LHE file %s", filename)

    if parse_events_as_xml:
        logger.debug("Parsing header and events as XML with %sElementTree", "c" if use_celementtree else "")
    else:
        logger.debug(
            "Parsing header as XML with %sElementTree and events as text file", "c" if use_celementtree else ""
        )

    # Inputs
    if cuts is None:
        cuts = OrderedDict()
    if efficiencies is None:
        efficiencies = OrderedDict()
    if k_factor is None:
        k_factor = 1.0
    if is_background and benchmark_names is None:
        raise RuntimeError("Parsing background LHE files required benchmark names to be provided.")

    # Unzip and open LHE file
    run_card = None
    for elem in _untar_and_parse_lhe_file(filename):
        if elem.tag == "MGRunCard":
            run_card = elem.text
            break
        else:
            continue

    # Figure out event weighting
    weight_norm_is_average = None
    n_events_runcard = None
    for line in run_card.splitlines():
        # Remove run card comments
        try:
            line, _ = line.split("!")
        except:
            pass

        # Separate in keys and values
        try:
            value, key = line.split("=")
        except:
            continue

        # Remove spaces
        value = value.strip()
        key = key.strip()

        # Parse entries
        if key == "nevents":
            n_events_runcard = float(value)
        if key == "event_norm":
            weight_norm_is_average = value == "average"

            logger.debug(
                "Found entry event_norm = %s in LHE header. Interpreting this as weight_norm_is_average " "= %s.",
                value,
                weight_norm_is_average,
            )

    if weight_norm_is_average is None:
        logger.warning(
            "Cannot read weight normalization mode (entry 'event_norm') from LHE file header. MadMiner "
            "will continue assuming that events are properly normalized. Please check this!"
        )

    # If necessary, rescale by number of events
    if weight_norm_is_average:
        if n_events_runcard is None:
            raise RuntimeError(
                "LHE weights have to be normalized, but MadMiner cannot read number of events (entry "
                "'nevents') from LHE file header."
            )

        k_factor = k_factor / n_events_runcard

    # Loop over events
    n_events_with_negative_weights = 0
    pass_cuts = [0 for _ in cuts]
    fail_cuts = [0 for _ in cuts]
    pass_efficiencies = [0 for _ in efficiencies]
    fail_efficiencies = [0 for _ in efficiencies]
    avg_efficiencies = [0 for _ in efficiencies]

    observations_all_events = []
    weights_all_events = []
    weight_names_all_events = None

    # Option one: XML parsing
    if parse_events_as_xml:
        events = _untar_and_parse_lhe_file(filename, ["event"])
        for i_event, event in enumerate(events, start=1):
            if i_event % 100000 == 0:
                logger.info("  Processing event %d/%d", i_event, n_events_runcard)

            # Parse event
            particles, weights, global_event_data = _parse_xml_event(event, sampling_benchmark)
            n_events_with_negative_weights, observations, pass_all, weight_names_all_events, weights = _parse_event(
                avg_efficiencies,
                cuts,
                efficiencies,
                energy_resolutions,
                eta_resolutions,
                fail_cuts,
                fail_efficiencies,
                n_events_with_negative_weights,
                observables,
                particles,
                pass_cuts,
                pass_efficiencies,
                phi_resolutions,
                pt_resolutions,
                weight_names_all_events,
                weights,
                global_event_data=global_event_data,
                print_event=i_event if i_event <= 20 else 0,
            )

            # Skip events that fail anything
            if not pass_all:
                continue

            # Store results
            observations_all_events.append(observations)
            weights_all_events.append(weights)

    # Option two: text parsing
    else:
        # Iterate over events in LHE file
        for i_event, (particles, weights) in enumerate(_parse_txt_events(filename, sampling_benchmark), start=1):
            if i_event % 100000 == 0:
                logger.info("  Processing event %d/%d", i_event, n_events_runcard)

            n_events_with_negative_weights, observations, pass_all, weight_names_all_events, weights = _parse_event(
                avg_efficiencies,
                cuts,
                efficiencies,
                energy_resolutions,
                eta_resolutions,
                fail_cuts,
                fail_efficiencies,
                n_events_with_negative_weights,
                observables,
                particles,
                pass_cuts,
                pass_efficiencies,
                phi_resolutions,
                pt_resolutions,
                weight_names_all_events,
                weights,
                print_event=i_event if i_event <= 20 else 0,
            )

            # Skip events that fail anything
            if not pass_all:
                continue

            # Store results
            observations_all_events.append(observations)
            weights_all_events.append(weights)

    # Check results
    n_events_pass = _report_parse_results(
        avg_efficiencies,
        cuts,
        efficiencies,
        fail_cuts,
        fail_efficiencies,
        n_events_with_negative_weights,
        observations_all_events,
        pass_cuts,
        pass_efficiencies,
    )

    if n_events_pass == 0:
        logger.warning("  No observations remaining!")
        return None, None

    # Reformat observations to OrderedDicts with entries {observable_name : (n_events,)}
    observations_all_events = list(map(list, zip(*observations_all_events)))  # transposes to (n_observables, n_events)
    observations_dict = OrderedDict()
    for key, values in zip(observables.keys(), observations_all_events):
        observations_dict[key] = np.asarray(values)

    # Reformat weights and add k-factors to weights
    weights_all_events = np.array(weights_all_events)  # (n_events, n_weights)
    weights_all_events = k_factor * weights_all_events
    weights_all_events = OrderedDict(zip(weight_names_all_events, weights_all_events.T))

    # Background events
    if is_background:
        for benchmark_name in benchmark_names:
            weights_all_events[benchmark_name] = weights_all_events[sampling_benchmark]

    # Re-organize weights again -- necessary for background events and nuisance benchmarks
    output_weights = OrderedDict()
    for benchmark_name in benchmark_names:
        if is_background:
            output_weights[benchmark_name] = weights_all_events[sampling_benchmark]
        else:
            output_weights[benchmark_name] = weights_all_events[benchmark_name]

    for syst_name, syst_data in systematics_dict.items():
        for (
            nuisance_param_name,
            ((nuisance_benchmark0, weight_name0), (nuisance_benchmark1, weight_name1), processing),
        ) in syst_data.items():
            # Store first benchmark associated with nuisance param
            if weight_name0 is None:
                weight_name0 = sampling_benchmark
            if processing is None:
                output_weights[nuisance_benchmark0] = weights_all_events[weight_name0]
            elif isinstance(processing, float):
                output_weights[nuisance_benchmark0] = processing * weights_all_events[weight_name0]
            else:
                raise RuntimeError(f"Unknown nuisance processing {processing}")

            # Store second benchmark associated with nuisance param
            if nuisance_benchmark1 is None or weight_name1 is None:
                continue
            if processing is None:
                output_weights[nuisance_benchmark1] = weights_all_events[weight_name1]
            elif isinstance(processing, float):
                output_weights[nuisance_benchmark1] = processing * weights_all_events[weight_name1]
            else:
                raise RuntimeError(f"Unknown nuisance processing {processing}")

    return observations_dict, output_weights


def _report_parse_results(
    avg_efficiencies,
    cuts,
    efficiencies,
    fail_cuts,
    fail_efficiencies,
    n_events_with_negative_weights,
    observations_all_events,
    pass_cuts,
    pass_efficiencies,
):
    for n_pass, n_fail, cut in zip(pass_cuts, fail_cuts, cuts):
        logger.info("  %s / %s events pass cut %s", n_pass, n_pass + n_fail, cut)
    for n_pass, n_fail, efficiency in zip(pass_efficiencies, fail_efficiencies, efficiencies):
        logger.info("  %s / %s events pass efficiency %s", n_pass, n_pass + n_fail, efficiency)
    for n_eff, efficiency, n_pass, n_fail in zip(avg_efficiencies, efficiencies, pass_efficiencies, fail_efficiencies):
        logger.info("  average efficiency for %s is %s", efficiency, n_eff / (n_pass + n_fail))

    n_events_pass = len(observations_all_events)

    if len(cuts) > 0:
        logger.info("  %s events pass all cuts/efficiencies", n_events_pass)
    if n_events_with_negative_weights > 0:
        logger.warning("  %s events contain negative weights", n_events_with_negative_weights)

    return n_events_pass


def _parse_event(
    avg_efficiencies,
    cuts: List[Cut],
    efficiencies: List[Efficiency],
    energy_resolutions,
    eta_resolutions,
    fail_cuts,
    fail_efficiencies,
    n_events_with_negative_weights,
    observables: Dict[str, Observable],
    particles,
    pass_cuts,
    pass_efficiencies,
    phi_resolutions,
    pt_resolutions,
    weight_names_all_events,
    weights,
    global_event_data=None,
    print_event=0,
):
    # Negative weights?
    n_events_with_negative_weights = _report_negative_weights(n_events_with_negative_weights, weights)
    if weight_names_all_events is None:
        weight_names_all_events = list(weights.keys())
    weights = np.array(list(weights.values()))

    # Apply smearing
    particles_smeared = _smear_particles(
        particles, energy_resolutions, pt_resolutions, eta_resolutions, phi_resolutions
    )

    # Objects in event
    try:
        variables = _get_objects(
            particles_smeared, particles, pt_resolutions["met"], global_event_data=global_event_data
        )
    except (TypeError, IndexError):
        variables = _get_objects(particles_smeared, particles, met_resolution=None, global_event_data=global_event_data)

    # Observables
    observations, pass_all_observation = _parse_observations(observables, variables)

    # Cuts
    pass_all_cuts = True
    if pass_all_observation:
        pass_all_cuts = _parse_cuts(
            cuts,
            fail_cuts,
            observables,
            observations,
            pass_all_cuts,
            pass_cuts,
            variables,
        )

    # Efficiencies
    pass_all_efficiencies = True
    if pass_all_observation and pass_all_cuts:
        pass_all_efficiencies, total_efficiency = _parse_efficiencies(
            avg_efficiencies,
            efficiencies,
            fail_efficiencies,
            pass_efficiencies,
            variables,
        )

        if pass_all_efficiencies:
            weights *= total_efficiency

    pass_all = pass_all_cuts and pass_all_efficiencies and pass_all_observation

    if print_event > 0:
        logger.debug(
            "Event {} {} observations, {} cuts, {} efficiencies -> {}".format(
                print_event,
                "passes" if pass_all_observation else "FAILS",
                "passes" if pass_all_cuts else "FAILS",
                "passes" if pass_all_efficiencies else "FAILS",
                "passes" if pass_all else "FAILS",
            )
        )
    return n_events_with_negative_weights, observations, pass_all, weight_names_all_events, weights


def _report_negative_weights(n_events_with_negative_weights, weights):
    n_negative_weights = np.sum(np.array(list(weights.values())) < 0.0)

    if n_negative_weights > 0:
        n_events_with_negative_weights += 1
        if n_events_with_negative_weights <= 3:
            logger.warning("Found %s negative weights in event. Weights: %s", n_negative_weights, weights)
        if n_events_with_negative_weights == 3:
            logger.warning("Skipping warnings about negative weights from now on...")

    return n_events_with_negative_weights


def _parse_observations(observables: Dict[str, Observable], variables: Dict[str, list]):
    observations = []
    passed_all = True

    for name, observable in observables.items():
        definition = observable.val_expression
        default = observable.val_default
        required = observable.is_required

        try:
            if isinstance(definition, str):
                value = eval(definition, variables)
            elif isinstance(definition, Callable):
                value = definition(
                    variables["p_truth"],
                    variables["l"],
                    variables["a"],
                    variables["j"],
                    variables["met"],
                )
            else:
                raise TypeError("Not a valid observable")
        except (IndexError, NameError, RuntimeError, SyntaxError, TypeError, ZeroDivisionError):
            passed_all = False if required else True
            value = default if default is not None else np.nan
        finally:
            observations.append(value)

    return observations, passed_all


def _parse_efficiencies(
    avg_efficiencies,
    efficiencies,
    fail_efficiencies,
    pass_efficiencies,
    variables,
):
    # Apply efficiencies
    total_efficiency = 1.0
    pass_all_efficiencies = True

    for i_efficiency, efficiency in enumerate(efficiencies):
        definition = efficiency.val_expression
        default = efficiency.val_default

        try:
            efficiency_result = eval(definition, variables)
            if efficiency_result > 0.0:
                pass_efficiencies[i_efficiency] += 1
                total_efficiency *= efficiency_result
                avg_efficiencies[i_efficiency] += efficiency_result
            else:
                fail_efficiencies[i_efficiency] += 1
                pass_all_efficiencies = False

        except (SyntaxError, NameError, TypeError, ZeroDivisionError, IndexError):
            if default > 0.0:
                pass_efficiencies[i_efficiency] += 1
                total_efficiency *= default
                avg_efficiencies[i_efficiency] += default
            else:
                fail_efficiencies[i_efficiency] += 1
                pass_all_efficiencies = False

    return pass_all_efficiencies, total_efficiency


def _parse_cuts(cuts, fail_cuts, observables, observations, pass_all_cuts, pass_cuts, variables):
    # Objects for cuts
    for obs_name, obs_value in zip(observables.keys(), observations):
        variables[obs_name] = obs_value

    # Check cuts
    for i_cut, cut in enumerate(cuts):
        definition = cut.val_expression
        required = cut.is_required

        try:
            cut_result = eval(definition, variables)
            if cut_result:
                pass_cuts[i_cut] += 1
            else:
                fail_cuts[i_cut] += 1
                pass_all_cuts = False

        except (SyntaxError, NameError, TypeError, ZeroDivisionError, IndexError):
            if required:
                pass_cuts[i_cut] += 1
            else:
                fail_cuts[i_cut] += 1
                pass_all_cuts = False

    return pass_all_cuts


def extract_nuisance_parameters_from_lhe_file(filename: str, systematics: Dict[str, Systematic]):
    """
    Extracts the definition of nuisance parameters from the LHE file
    and returns a systematics_dict with structure:
    {systematics_name: {nuisance_parameter_name: ((benchmark0, weight0), (benchmark1, weight1), processing) }
    """

    logger.debug("Parsing nuisance parameter setup from LHE file at %s", filename)

    # Nuisance parameters (output)
    systematics_dict = OrderedDict()

    # When no systematics setup is defined
    if systematics is None:
        return systematics_dict

    # Parse scale factors from strings in systematics
    logger.debug("Systematics setup: %s", systematics)

    # Unzip and parse LHE file
    initrwgts = _untar_and_parse_lhe_file(filename, ["initrwgt"])

    # Find weight groups
    weight_groups = []
    try:
        for initrwgt in initrwgts:
            weight_groups += initrwgt.findall("weightgroup")
    except KeyError as e:
        raise RuntimeError("Could not find weight groups in LHE file!\n%s", e)
    # if len(weight_groups) == 0:
    #     raise RuntimeError("Zero weight groups in LHE file!")
    logger.debug("%s weight groups", len(weight_groups))

    # Loop over systematics
    for syst_name, syst_obj in systematics.items():
        nuisance_param_dict = _extract_nuisance_param_dict(weight_groups, syst_name, syst_obj)
        systematics_dict[syst_name] = nuisance_param_dict

    return systematics_dict


def _extract_nuisance_param_dict(weight_groups: list, systematics_name: str, systematic: Systematic):
    logger.debug("Extracting nuisance parameter information for systematic %s", systematics_name)

    if systematic.type == SystematicType.NORM:
        nuisance_param_name = f"{systematics_name}_nuisance_param_0"
        benchmark_name = f"{nuisance_param_name}_benchmark_0"
        nuisance_param_definition = (benchmark_name, None), (None, None), systematic.value
        return {nuisance_param_name: nuisance_param_definition}

    elif systematic.type == SystematicType.SCALE:
        # Prepare output
        nuisance_param_definition_parts = []

        # Parse scale variations we need to find
        scale_factors = systematic.value.split(",")
        scale_factors = [float(sf) for sf in scale_factors]

        if len(scale_factors) == 0:
            raise RuntimeError("Cannot parse scale factor string %s", systematic.value)
        elif len(scale_factors) == 1:
            scale_factors = (scale_factors[0],)
        else:
            scale_factors = (scale_factors[-1], scale_factors[0])

        # Loop over scale factors
        for k, scale_factor in enumerate(scale_factors):
            muf = scale_factor if systematic.scale in {SystematicScale.MU, SystematicScale.MUF} else 1.0
            mur = scale_factor if systematic.scale in {SystematicScale.MU, SystematicScale.MUR} else 1.0

            # Loop over weight groups and weights and identify benchmarks
            for wg in weight_groups:
                try:
                    wg_name = wg.attrib["name"]
                except KeyError:
                    logger.warning("New weight group: does not have name attribute, skipping")
                    continue
                logger.debug("New weight group: %s", wg_name)

                if "mg_reweighting" in wg_name.lower() or "scale variation" not in wg_name.lower():
                    continue
                logger.debug("Weight group identified as scale variation")

                weights = wg.findall("weight")

                for weight in weights:
                    try:
                        weight_id = str(weight.attrib["id"])
                        weight_muf = float(weight.attrib["MUF"])
                        weight_mur = float(weight.attrib["MUR"])
                    except KeyError:
                        logger.warning("Scale variation weight does not have all expected attributes")
                        continue

                    logging.debug(
                        "Found scale variation weight %s / muf = %s, mur = %s", weight_id, weight_muf, weight_mur
                    )

                    # Let's skip the entries with a varied dynamical scale for now
                    weight_dynscale = None
                    for key in ["dynscale", "dyn_scale", "DYNSCALE", "DYN_SCALE"]:
                        try:
                            weight_dynscale = int(weight.attrib[key])
                        except KeyError:
                            pass
                    if weight_dynscale is not None:
                        continue

                    # Matching time!
                    if approx_equal(weight_mur, mur) and approx_equal(weight_muf, muf):
                        benchmark_name = f"{systematics_name}_nuisance_param_0_benchmark_{k}"
                        nuisance_param_definition_parts.append((benchmark_name, weight_id))
                        break

        if len(nuisance_param_definition_parts) < len(scale_factors):
            logger.warning(
                "Could not find weights for the scale uncertainty %s in LHE file! The most common source of "
                " this error is not having installed LHAPDF with its Python interface. Please make sure that"
                " you have installed this. You can also check the log file produced by MadGraph for a "
                "warning about this. If LHAPDF is correctly installed and you still get this warning, please"
                " check manually whether the LHE file contains weights from PDF variation, and contact"
                " the MadMiner developer team about this. If you continue with the analysis, MadMiner"
                " will disregard PDF uncertainties.",
                systematics_name,
            )
            return {}
        else:
            # Output
            nuisance_param_name = f"{systematics_name}_nuisance_param_0"
            if len(nuisance_param_definition_parts) > 1:
                nuisance_dict = {
                    nuisance_param_name: (nuisance_param_definition_parts[0], nuisance_param_definition_parts[1], None)
                }
            else:
                nuisance_dict = {nuisance_param_name: (nuisance_param_definition_parts[0], (None, None), None)}
            return nuisance_dict

    elif systematic.type == SystematicType.PDF:
        nuisance_dict = OrderedDict()

        # Loop over weight groups and weights and identify benchmarks
        for wg in weight_groups:
            try:
                wg_name = wg.attrib["name"]
            except KeyError:
                logger.warning("New wWeight group: does not have name attribute, skipping")
                continue
            logger.debug("New weight group: %s", wg_name)

            if (
                "mg_reweighting" in wg_name.lower()
                or "pdf" not in wg_name.lower()
                or "ct" not in wg_name.lower()
                or systematic.value not in wg_name.lower()
            ):
                continue

            logger.debug("Weight group identified as PDF variation")
            weights = wg.findall("weight")

            for i, weight in enumerate(weights):
                try:
                    weight_id = str(weight.attrib["id"])
                    weight_pdf = int(weight.attrib["PDF"])
                except KeyError:
                    logger.warning("Scale variation weight does not have all expected attributes")
                    continue

                if weight_pdf % 1000 == 0:  # Central element, not eigenvector of covariance matrix
                    logger.debug("Identifying PDF weight %s / %s as central element", weight_id, weight_pdf)
                    continue

                logger.debug("Found PDF weight %s / %s", weight_id, weight_pdf)

                # Add every PDF Hessian direction to nuisance parameters
                nuisance_param_name = f"{systematics_name}_nuisance_param_{i}"
                benchmark_name = f"{nuisance_param_name}_benchmark_0"
                nuisance_dict[nuisance_param_name] = (benchmark_name, weight_id), (None, None), None

        # Check that everything was found
        if len(nuisance_dict) < 0:
            logger.warning(
                "Could not find weights for the PDF uncertainties in LHE file! The most common source of this"
                " error is not having installed LHAPDF with its Python interface. Please make sure that you "
                " have installed this. You can also check the log file produced by MadGraph for a warning"
                " about this. If LHAPDF is correctly installed and you still get this warning, please check"
                " manually whether the LHE file at %s contains weights from PDF variation, and contact"
                " the MadMiner developer team about this. If you continue with the analysis, MadMiner"
                " will disregard PDF uncertainties."
            )
        return nuisance_dict

    else:
        raise RuntimeError("Unknown systematics type %s", systematic.type)


def _parse_xml_event(event, sampling_benchmark):
    # Initialize weights and momenta
    weights = OrderedDict()
    particles = []
    global_event_data = {}

    # Split kinematics part in tag line and momenta
    event_text = event.text
    tag_line = None
    particle_lines = []
    for line in event_text.splitlines():
        elements = line.split()
        if len(elements) < 2:
            continue
        if tag_line is None:
            tag_line = elements
        else:
            particle_lines.append(elements)

    # Parse tag
    assert tag_line is not None
    global_event_data["n_particles"] = float(tag_line[0])
    weights[sampling_benchmark] = float(tag_line[2])
    global_event_data["scale"] = float(tag_line[3])
    global_event_data["alpha_qed"] = float(tag_line[4])
    global_event_data["alpha_qcd"] = float(tag_line[5])

    # Parse momenta
    for elements in particle_lines:
        if len(elements) < 10:
            continue
        status = int(elements[1])
        if elements[0] == "#aMCatNLO":
            elements = elements[1:]
        if status == 1:
            pdgid = int(elements[0])
            px = float(elements[6])
            py = float(elements[7])
            pz = float(elements[8])
            e = float(elements[9])
            spin = float(elements[12])
            particle = MadMinerParticle.from_xyzt(px, py, pz, e)
            particle.set_pdgid(pdgid)
            particle.set_spin(spin)
            particles.append(particle)

    # Weights
    if event.find("rwgt") is not None:
        for weight in event.find("rwgt").findall("wgt"):
            weight_id, weight_value = weight.attrib["id"], float(weight.text)
            weights[weight_id] = weight_value

    return particles, weights, global_event_data


def _parse_txt_events(filename, sampling_benchmark):
    # Initialize weights and momenta
    weights = OrderedDict()
    particles = []

    # Some tags so that we know where in the event we are
    do_tag = False
    do_momenta = False
    do_reweight = False

    # Event
    n_events = 0

    reset_event = False

    # Loop through lines in Event
    with open(filename, "r") as file:
        for line in file.readlines():
            # Clean up line
            try:
                line = line.split("#")[0]
            except:
                pass
            line = line.strip()
            elements = line.split()

            # Skip empty/commented out lines
            if len(line) == 0 or len(elements) == 0:
                continue

            # End of LHE file
            elif line == "</LesHouchesEvents>":
                return

            # Beginning of event
            elif line == "<event>":
                # Initialize weights and momenta
                weights = OrderedDict()
                particles = []

                # Some tags so that we know where in the event we are
                do_tag = True
                do_momenta = False
                do_reweight = False

            # End of event
            elif line == "</event>":
                n_events += 1
                yield particles, weights

                # Reset weights and momenta
                weights = OrderedDict()
                particles = []

                # Some tags so that we know where in the event we are
                do_tag = False
                do_momenta = False
                do_reweight = False

            # Beginning of unimportant block
            elif line == "<mgrwt>":
                do_tag = False
                do_momenta = False
                do_reweight = False

            # Beginning of weight block
            elif line == "<rwgt>":
                do_tag = False
                do_momenta = False
                do_reweight = True

            # End of weight block
            elif line == "</rwgt>":
                do_tag = False
                do_momenta = False
                do_reweight = False

            # Read tag -> first weight
            elif do_tag:
                weights[sampling_benchmark] = float(elements[2])

                do_tag = False
                do_momenta = True
                do_reweight = False

            # Read Momenta and store as 4-vector
            elif do_momenta:
                status = int(elements[1])
                if status == 1:
                    pdgid = int(elements[0])
                    px = float(elements[6])
                    py = float(elements[7])
                    pz = float(elements[8])
                    e = float(elements[9])
                    spin = float(elements[12])
                    particle = MadMinerParticle.from_xyzt(px, py, pz, e)
                    particle.set_pdgid(pdgid)
                    particle.set_spin(spin)
                    particles.append(particle)

            # Read reweighted weights
            elif do_reweight:
                rwgtid = line[line.find("<") + 1 : line.find(">")].split("=")[1][1:-1]
                rwgtval = float(line[line.find(">") + 1 : line.find("<", line.find("<") + 1)])
                weights[rwgtid] = rwgtval


def _parse_lhe_file_with_bad_chars(filename):
    # In some cases, the LHE comments can contain bad characters
    with open(filename, "r") as file:
        for line in file:
            comment_pos = line.find("#")
            if comment_pos >= 0:
                yield line[:comment_pos]
            else:
                yield line


def _untar_and_parse_lhe_file(filename, tags=None):
    # Unzip event file
    new_filename, extension = os.path.splitext(filename)
    if extension == ".gz":
        if not os.path.exists(new_filename):
            unzip_file(filename, new_filename)
        filename = new_filename

    for event, elem in ET.iterparse(filename):
        if tags and elem.tag not in tags:
            continue
        else:
            yield elem

        elem.clear()


def _get_objects(particles, particles_truth, met_resolution=None, global_event_data=None):
    # Find visible particles
    electrons = []
    muons = []
    taus = []
    photons = []
    jets = []
    leptons = []
    neutrinos = []
    unstables = []
    invisibles = []

    for particle in particles:
        pdgid = abs(particle.pdgid)
        if pdgid in [1, 2, 3, 4, 5, 6, 9, 21]:
            if pdgid == 5:
                particle.set_tags(False, True, False)
            if pdgid == 6:
                particle.set_tags(False, False, True)
            jets.append(particle)
        elif pdgid == 11:
            electrons.append(particle)
            leptons.append(particle)
        elif pdgid == 13:
            muons.append(particle)
            leptons.append(particle)
        elif pdgid == 15:
            taus.append(particle)
        elif pdgid == 22:
            photons.append(particle)
        elif pdgid in [12, 14, 16]:
            neutrinos.append(particle)
            invisibles.append(particle)
        elif pdgid in [23, 24, 25]:
            unstables.append(particle)
        else:
            logger.warning("Unknown particle with PDG id %s, treating as invisible!")
            invisibles.append(particle)

    # Sort by pT
    electrons = sorted(electrons, reverse=True, key=lambda x: x.pt)
    muons = sorted(muons, reverse=True, key=lambda x: x.pt)
    taus = sorted(taus, reverse=True, key=lambda x: x.pt)
    photons = sorted(photons, reverse=True, key=lambda x: x.pt)
    leptons = sorted(leptons, reverse=True, key=lambda x: x.pt)
    neutrinos = sorted(neutrinos, reverse=True, key=lambda x: x.pt)
    jets = sorted(jets, reverse=True, key=lambda x: x.pt)

    # Sum over all particles
    ht = 0.0
    visible_sum = MadMinerParticle.from_xyzt(0.0, 0.0, 0.0, 0.0)

    standard_ids = set(get_elementary_pdg_ids())
    neutrino_ids = set(
        int(p.pdgid)
        for p in Particle.findall(lambda p: p.pdgid.is_sm_lepton and p.charge == 0)
    )

    for particle in particles:
        if particle.pdgid in standard_ids and particle.pdgid not in neutrino_ids:
            visible_sum += particle
            ht += particle.pt

    # Soft noise
    if met_resolution is not None:
        noise_std = met_resolution[0] + met_resolution[1] * ht
        noise_x = np.random.normal(0.0, noise_std, size=None)
        noise_y = np.random.normal(0.0, noise_std, size=None)
    else:
        noise_x = 0.0
        noise_y = 0.0

    # MET
    met_x = -visible_sum.x + noise_x
    met_y = -visible_sum.y + noise_y
    met = MadMinerParticle.from_xyzt(
        x=met_x,
        y=met_y,
        z=0.0,
        t=(met_x ** 2 + met_y ** 2) ** 0.5,
    )

    # Build objects
    objects = math_commands()
    objects.update(
        {
            "p": particles,
            "p_truth": particles_truth,
            "e": electrons,
            "j": jets,
            "a": photons,
            "mu": muons,
            "tau": taus,
            "l": leptons,
            "met": met,
            "v": neutrinos,
        }
    )

    # Global event_data
    if global_event_data is not None:
        objects.update(global_event_data)

    return objects


def _smear_variable(true_value, resolutions, id):
    """ Adds Gaussian nose to a variable """
    try:
        res = float(resolutions[id][0] + resolutions[id][1] * true_value)

        if res <= 0.0:
            return true_value

        return float(true_value + np.random.normal(0.0, float(res), None))

    except (TypeError, KeyError):
        return true_value


def _smear_particles(particles, energy_resolutions, pt_resolutions, eta_resolutions, phi_resolutions):
    """ Applies smearing function to particles of one event """

    # No smearing if any argument is None
    if energy_resolutions is None or pt_resolutions is None or eta_resolutions is None or phi_resolutions is None:
        return particles

    smeared_particles = []

    for particle in particles:
        pdgid = particle.pdgid

        if (
            pdgid not in energy_resolutions.keys()
            or pdgid not in pt_resolutions.keys()
            or pdgid not in eta_resolutions.keys()
            or pdgid not in phi_resolutions.keys()
        ):
            continue

        if None in energy_resolutions[pdgid] and None in pt_resolutions[pdgid]:
            raise RuntimeError("Cannot derive both pT and energy from on-shell conditions!")

        # Minimum energy and pT
        m = particle.m
        min_e = 0.0
        if None in pt_resolutions[pdgid]:
            min_e = m
        min_pt = 0.0

        # Smear four-momenta
        e = None
        if None not in energy_resolutions[pdgid]:
            e = min_e - 1.0
            while e <= min_e:
                e = _smear_variable(particle.e, energy_resolutions, pdgid)
        pt = None
        if None not in pt_resolutions[pdgid]:
            pt = min_pt - 1.0
            while pt <= min_pt:
                pt = _smear_variable(particle.pt, pt_resolutions, pdgid)
        eta = _smear_variable(particle.eta, eta_resolutions, pdgid)
        phi = _smear_variable(particle.phi, phi_resolutions, pdgid)
        while phi > 2.0 * np.pi:
            phi -= 2.0 * np.pi
        while phi < 0.0:
            phi += 2.0 * np.pi

        if None in energy_resolutions[pdgid]:
            # Calculate E from on-shell conditions
            smeared_particle = MadMinerParticle.from_rhophietatau(pt, phi, eta, particle.m)

        elif None in pt_resolutions[pdgid]:
            # Calculate pT from on-shell conditions
            if e > m:
                pt = (e ** 2 - m ** 2) ** 0.5 / np.cosh(eta)
            else:
                pt = 0.0
            smeared_particle = MadMinerParticle.from_rhophietat(pt, phi, eta, e)

        else:
            # Everything smeared manually
            smeared_particle = MadMinerParticle.from_rhophietat(pt, phi, eta, e)

        # PDG id (also sets charge)
        smeared_particle.set_pdgid(pdgid)
        smeared_particles.append(smeared_particle)

    return smeared_particles


def get_elementary_pdg_ids():
    """Get Standard Model elementary particle IDs"""
    pdg_ids = Particle.findall(
        lambda p: (
            p.pdgid.is_sm_quark
            or p.pdgid.is_sm_lepton
            or p.pdgid.is_sm_gauge_boson_or_higgs
        )
    )

    return [int(p.pdgid) for p in pdg_ids]



import logging
import os
import shutil

from madminer.utils.various import call_command, make_file_executable, create_missing_folders

logger = logging.getLogger(__name__)


def generate_mg_process(
    mg_directory,
    temp_directory,
    proc_card_file,
    mg_process_directory,
    ufo_model_directory=None,
    log_file=None,
    initial_command=None,
    python_executable=None,
):

    """
    Calls MadGraph to create the process folder.

    Parameters
    ----------
    mg_directory : str
        Path to the MadGraph 5 directory.

    temp_directory : str
        Path to a directory for temporary files.

    proc_card_file : str
        Path to the process card that tells MadGraph how to generate the process.

    mg_process_directory : str
        Path to the MG process directory.

    ufo_model_directory : str or None, optional
        Path to a UFO model that is not yet installed. It will be copied to the MG directory before the process card
        is executed. Default value: None.

    initial_command : str or None, optional
        Initial bash commands that have to be executed before MG is run (e.g. to load the correct virtual
        environment). Default value: None.

    log_file : str or None, optional
        Path to a log file in which the MadGraph output is saved. Default value: None.

    python_executable : None or str, optional
        Overwrites the default Python executable

    Returns
    -------
        None

    """

    # Preparations
    logger.info("Generating MadGraph process folder from %s at %s", proc_card_file, mg_process_directory)

    create_missing_folders([temp_directory, mg_process_directory, os.path.dirname(log_file)])

    if ufo_model_directory is not None:
        copy_ufo_model(ufo_model_directory, mg_directory)

    # MG commands
    temp_proc_card_file = f"{temp_directory}/generate.mg5"
    shutil.copyfile(proc_card_file, temp_proc_card_file)

    with open(temp_proc_card_file, "a") as myfile:
        myfile.write(f"\n")
        myfile.write(f"\n")
        myfile.write(f"output {mg_process_directory}")

    # Call specific initial command and Python binary
    initial_command = f"{initial_command}; " if initial_command is not None else ""
    python_binary = f"{python_executable} " if python_executable is not None else ""

    command = f"{initial_command}{python_binary}{mg_directory}/bin/mg5_aMC {temp_proc_card_file}"
    logger.info(f"Calling MadGraph: {command}")

    _ = call_command(cmd=command, log_file=log_file)


def setup_mg_with_scripts(
    mg_process_directory,
    proc_card_filename_from_mgprocdir=None,
    run_card_file_from_mgprocdir=None,
    param_card_file_from_mgprocdir=None,
    reweight_card_file_from_mgprocdir=None,
    pythia8_card_file_from_mgprocdir=None,
    configuration_file_from_mgprocdir=None,
    is_background=False,
    script_file_from_mgprocdir=None,
    initial_command=None,
    log_dir=None,
    log_file_from_logdir=None,
    order="LO",
    python_executable=None,
):
    """
    Prepares a bash script that will start the event generation.

    Parameters
    ----------
    mg_process_directory : str
        Path to the MG process directory.

    proc_card_filename_from_mgprocdir : str or None, optional
        Filename for the MG command card that will be generated, relative from mg_process_directory. If None, a
        default filename in the MG process directory will be chosen.

    run_card_file_from_mgprocdir : str or None, optional
        Path to the MadGraph run card, relative from mg_process_directory. If None, the card present in the process
        folder is used. Default value: None.

    param_card_file_from_mgprocdir : str or None, optional
        Path to the MadGraph run card, relative from mg_process_directory. If None, the card present in the process
        folder is used. Default value: None.

    reweight_card_file_from_mgprocdir : str or None, optional
        Path to the MadGraph reweight card, relative from mg_process_directory. If None, the card present in the
        process folder is used. Default value: None.

    pythia8_card_file_from_mgprocdir : str or None, optional
        Path to the MadGraph Pythia8 card, relative from mg_process_directory. If None, Pythia is not run. Default
        value: None.

    configuration_file_from_mgprocdir : str or None, optional
        Path to the MadGraph me5_configuration card, relative from mg_process_directory. If None, the card
        present in the process folder is used. Default value: None.

    is_background : bool, optional
        Should be True for background processes, i.e. process in which the differential cross section does not
        depend on the parameters (and would be the same for all benchmarks). In this case, no reweighting is run,
        which can substantially speed up the event generation. Default value: False.

    script_file_from_mgprocdir : str or None, optional
        This sets where the shell script to run MG and Pythia is generated, relative from mg_process_directory. If
        None, a default filename in `mg_process_directory/madminer` is used. Default value: None.

    initial_command : str or None, optional
        Initial shell commands that have to be executed before MG is run (e.g. to load a virtual environment).
        Default value: None.

    log_dir : str or None, optional
        Log directory. Default value: None.

    log_file_from_logdir : str or None, optional
        Path to a log file in which the MadGraph output is saved, relative from the default log directory. Default
        value: None.

    python_executable : None or str, optional
        Overwrites the default Python executable

    Returns
    -------
    bash_script_call : str
        How to call this script.

    """

    # Preparations
    create_missing_folders([mg_process_directory])
    if log_dir is not None:
        create_missing_folders([log_dir])
    if proc_card_filename_from_mgprocdir is not None:
        create_missing_folders([os.path.dirname(mg_process_directory + "/" + proc_card_filename_from_mgprocdir)])

    # Prepare run...
    logger.info("Preparing script to run MadGraph and Pythia in %s", mg_process_directory)

    # Bash script can optionally provide MG path or process directory
    mg_directory_placeholder = "$mgdir"
    mg_process_directory_placeholder = "$mgprocdir"
    log_dir_placeholder = "$mmlogdir"
    placeholder_definition = "mgdir=$1\nmgprocdir=$2\nmmlogdir=$3"

    # Find filenames for process card and script
    if proc_card_filename_from_mgprocdir is None:
        for i in range(1000):
            proc_card_filename_from_mgprocdir = f"/Cards/start_event_generation_{i}.mg5"
            if not os.path.isfile(f"{mg_process_directory}/{proc_card_filename_from_mgprocdir}"):
                break
    else:
        proc_card_filename = mg_process_directory + "/" + proc_card_filename_from_mgprocdir

    if script_file_from_mgprocdir is None:
        for i in range(1000):
            script_file = f"{mg_process_directory}/madminer/scripts/madminer_run_{i}.sh"
            if not os.path.isfile(script_file):
                break
    else:
        script_file = f"{mg_process_directory}/{script_file_from_mgprocdir}"

    script_filename = os.path.basename(script_file)

    if log_file_from_logdir is None:
        log_file_from_logdir = "/log.log"

    # MG commands
    shower_option = "OFF" if pythia8_card_file_from_mgprocdir is None else "Pythia8"
    reweight_option = "OFF" if is_background else "ON"

    mg_commands = """
        launch {}
        shower={}
        detector=OFF
        analysis=OFF
        madspin=OFF
        reweight={}
        done
        """.format(
        mg_process_directory_placeholder, shower_option, reweight_option
    )

    with open(proc_card_filename, "w") as file:
        file.write(mg_commands)

    #  Card copying commands
    copy_commands = ""
    if run_card_file_from_mgprocdir is not None:
        copy_commands += (
            f"cp "
            f"{mg_process_directory_placeholder}/{run_card_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/run_card.dat\n"
        )
    if param_card_file_from_mgprocdir is not None:
        copy_commands += (
            f"cp "
            f"{mg_process_directory_placeholder}/{param_card_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/param_card.dat\n"
        )
    if reweight_card_file_from_mgprocdir is not None and not is_background:
        copy_commands += (
            f"cp "
            f"{mg_process_directory_placeholder}/{reweight_card_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/reweight_card.dat\n"
        )
    if pythia8_card_file_from_mgprocdir is not None and order == "LO":
        copy_commands += (
            f"cp "
            f"{mg_process_directory_placeholder}/{pythia8_card_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/pythia8_card.dat\n"
        )
    if pythia8_card_file_from_mgprocdir is not None and order == "NLO":
        copy_commands += (
            f"cp "
            f"{mg_process_directory_placeholder}/{pythia8_card_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/shower_card.dat\n"
        )
    if configuration_file_from_mgprocdir is not None:
        copy_commands += (
            f"cp "
            f"{mg_process_directory_placeholder}/{configuration_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/me5_configuration.txt\n"
        )

    # Replace environment variable in proc card
    replacement_command = """sed -e 's@\$mgprocdir@'"$mgprocdir"'@' {}/{} > {}/{}""".format(
        mg_process_directory_placeholder,
        proc_card_filename_from_mgprocdir,
        mg_process_directory_placeholder,
        "Cards/mg_commands.mg5",
    )

    # Call specific initial command and Python binary
    initial_command = f"{initial_command} " if initial_command is not None else ""
    python_binary = f"{python_executable} " if python_executable is not None else ""

    # Put together script
    script = (
        "#!/bin/bash\n\n# Script generated by MadMiner\n\n# Usage: {} MG_directory MG_process_directory log_dir\n\n"
        + "{}\n\n{}\n\n{}\n{}\n\n{} {}/bin/mg5_aMC {}/{} > {}/{}\n"
    ).format(
        script_filename,
        initial_command,
        placeholder_definition,
        copy_commands,
        replacement_command,
        python_binary,
        mg_directory_placeholder,
        mg_process_directory_placeholder,
        "Cards/mg_commands.mg5",
        log_dir_placeholder,
        log_file_from_logdir,
    )

    with open(script_file, "w") as file:
        file.write(script)
    make_file_executable(script_file)

    # How to call it from master script
    call_placeholder = (
        f"{mg_process_directory_placeholder}/{script_file_from_mgprocdir} "
        f"{mg_directory_placeholder} "
        f"{mg_process_directory_placeholder} "
        f"{log_dir_placeholder}"
    )

    return call_placeholder


def run_mg(
    mg_directory,
    mg_process_directory,
    proc_card_filename=None,
    run_card_file=None,
    param_card_file=None,
    reweight_card_file=None,
    pythia8_card_file=None,
    configuration_card_file=None,
    is_background=False,
    initial_command=None,
    log_file=None,
    order="LO",
    python_executable=None,
):
    """
    Calls MadGraph to generate events.

    Parameters
    ----------
    mg_directory : str
        Path to the MadGraph 5 base directory.

    mg_process_directory : str
        Path to the MG process directory.

    proc_card_filename : str or None, optional
        Filename for the MG command card that will be generated. If None, a default filename in the MG process
        directory will be chosen.

    run_card_file : str or None, optional
        Path to the MadGraph run card. If None, the card present in the process folder is used. Default value:
        None)

    param_card_file : str or None, optional
        Path to the MadGraph param card. If None, the card present in the process folder is used. Default value:
        None)

    reweight_card_file : str or None, optional
        Path to the MadGraph reweight card. If None, the card present in the process folder is used. (Default value
        = None)

    pythia8_card_file : str or None, optional
        Path to the MadGraph Pythia8 card. If None, Pythia is not run. Default value: None.

    configuration_card_file : str or None, optional
        Path to the MadGraph configuration card. If None, the card present in the process folder is used. (Default
        value: None).

    is_background : bool, optional
        Should be True for background processes, i.e. process in which the differential cross section does not
        depend on the parameters (and would be the same for all benchmarks). In this case, no reweighting is run,
        which can substantially speed up the event generation. Default value: False.

    initial_command : str or None, optional
        Initial shell commands that have to be executed before MG is run (e.g. to load a virtual environment).
        Default value: None.

    log_file : str or None, optional
        Path to a log file in which the MadGraph output is saved. Default value: None.

    python_executable : None or str, optional
        Overwrites the default Python executable

    Returns
    -------
        None

    """

    # Preparations
    create_missing_folders([mg_process_directory, os.path.dirname(log_file)])
    if proc_card_filename is not None:
        create_missing_folders([os.path.dirname(proc_card_filename)])

    # Just run it already
    logger.info("Starting MadGraph and Pythia in %s", mg_process_directory)

    # Copy cards
    if run_card_file is not None:
        shutil.copyfile(run_card_file, f"{mg_process_directory}/Cards/run_card.dat")
    if param_card_file is not None:
        shutil.copyfile(param_card_file, f"{mg_process_directory}/Cards/param_card.dat")
    if reweight_card_file is not None and not is_background:
        shutil.copyfile(reweight_card_file, f"{mg_process_directory}/Cards/reweight_card.dat")
    if pythia8_card_file is not None and order == "LO":
        shutil.copyfile(pythia8_card_file, f"{mg_process_directory}/Cards/pythia8_card.dat")
    if pythia8_card_file is not None and order == "NLO":
        shutil.copyfile(pythia8_card_file, f"{mg_process_directory}/Cards/shower_card.dat")
    if configuration_card_file is not None:
        shutil.copyfile(configuration_card_file, f"{mg_process_directory}/Cards/me5_configuration.txt")

    # Find filenames for process card and script
    if proc_card_filename is None:
        for i in range(1000):
            proc_card_filename = f"{mg_process_directory}/Cards/start_event_generation_{i}.mg5"
            if not os.path.isfile(proc_card_filename):
                break

    # MG commands
    shower_option = "OFF" if pythia8_card_file is None else "Pythia8"
    reweight_option = "OFF" if is_background else "ON"

    mg_commands = """
        launch {}
        shower={}
        detector=OFF
        analysis=OFF
        madspin=OFF
        reweight={}
        done
        """.format(
        mg_process_directory, shower_option, reweight_option
    )

    with open(proc_card_filename, "w") as file:
        file.write(mg_commands)

    # Call specific initial command and Python binary
    initial_command = f"{initial_command}; " if initial_command is not None else ""
    python_binary = f"{python_executable} " if python_executable is not None else ""

    command = f"{initial_command}{python_binary}{mg_directory}/bin/mg5_aMC {proc_card_filename}"
    logger.info(f"Calling MadGraph: {command}")

    _ = call_command(cmd=command, log_file=log_file)


def setup_mg_reweighting_with_scripts(
    mg_process_directory,
    run_name,
    reweight_card_file_from_mgprocdir=None,
    script_file_from_mgprocdir=None,
    initial_command=None,
    log_dir=None,
    log_file_from_logdir=None,
):
    """
    Prepares a bash script that will start the event generation.

    Parameters
    ----------
    mg_process_directory : str
        Path to the MG process directory.

    reweight_card_file_from_mgprocdir : str or None, optional
        Path to the MadGraph reweight card, relative from mg_process_directory. If None, the card present in the
        process folder is used. Default value: None.

    script_file_from_mgprocdir : str or None, optional
        This sets where the shell script to run the reweighting is generated, relative from mg_process_directory. If
        None, a default filename in `mg_process_directory/madminer` is used. Default value: None.

    initial_command : str or None, optional
        Initial shell commands that have to be executed before MG is run (e.g. to load a virtual environment).
        Default value: None.

    log_dir : str or None, optional
        Log directory. Default value: None.

    log_file_from_logdir : str or None, optional
        Path to a log file in which the MadGraph output is saved, relative from the default log directory. Default
        value: None.

    Returns
    -------
    bash_script_call : str
        How to call this script.

    """

    # Preparations
    if log_dir is not None:
        create_missing_folders([log_dir])

    # Prepare run...
    logger.info("Preparing script to reweight an existing sample in %s", mg_process_directory)

    # Bash script can optionally provide MG path or process directory
    mg_process_directory_placeholder = "$mgprocdir"
    log_dir_placeholder = "$mmlogdir"
    placeholder_definition = r"mgprocdir=${1:-" + mg_process_directory + r"}" + "\n"
    placeholder_definition += r"mmlogdir=${2:-" + log_dir + r"}"

    if script_file_from_mgprocdir is None:
        script_file = f"{mg_process_directory}/madminer/scripts/madminer_reweight_{run_name}.sh"
    else:
        script_file = f"{mg_process_directory}/{script_file_from_mgprocdir}"

    script_filename = os.path.basename(script_file)

    if log_file_from_logdir is None:
        log_file_from_logdir = "/log.log"

    # Initial commands
    if initial_command is None:
        initial_command = ""

    #  Card copying commands
    if reweight_card_file_from_mgprocdir is not None:
        copy_commands = (
            f"cp "
            f"{mg_process_directory_placeholder}/{reweight_card_file_from_mgprocdir} "
            f"{mg_process_directory_placeholder}/Cards/reweight_card.dat\n"
        )
    else:
        copy_commands = ""

    # Put together script
    script = (
        "#!/bin/bash\n\n# Script generated by MadMiner\n\n# Usage: {} MG_process_directory log_dir\n\n"
        + "{}\n\n{}\n\n{}\n\n{}/bin/madevent reweight {} -f > {}/{}\n"
    ).format(
        script_filename,
        initial_command,
        placeholder_definition,
        copy_commands,
        mg_process_directory_placeholder,
        run_name,
        log_dir_placeholder,
        log_file_from_logdir,
    )

    with open(script_file, "w") as file:
        file.write(script)
    make_file_executable(script_file)

    # How to call it from master script
    call_instruction = (
        f"{mg_process_directory}/{script_file_from_mgprocdir} "
        f"[MG_process_directory] [log_directory]"
    )

    return call_instruction


def run_mg_reweighting(mg_process_directory, run_name, reweight_card_file=None, initial_command=None, log_file=None):
    """
    Runs MG reweighting.

    Parameters
    ----------
    mg_process_directory : str
        Path to the MG process directory.

    run_name : str
        Run name.

    reweight_card_file : str or None, optional
        Path to the MadGraph reweight card. If None, the card present in the process folder is used. (Default value
        = None)

    initial_command : str or None, optional
        Initial shell commands that have to be executed before MG is run (e.g. to load a virtual environment).
        Default value: None.

    log_file : str or None, optional
        Path to a log file in which the MadGraph output is saved. Default value: None.

    Returns
    -------
    bash_script_call : str
        How to call this script.

    """

    # Preparations
    create_missing_folders([os.path.dirname(log_file)])

    # Prepare run...
    logger.info("Starting reweighting of an existing sample in %s", mg_process_directory)

    #  Copy cards
    if reweight_card_file is not None:
        shutil.copyfile(reweight_card_file, mg_process_directory + "/Cards/reweight_card.dat")

    # Call MG5 reweight feature
    initial_command = f"{initial_command}; " if initial_command else ""

    _ = call_command(
        cmd=f"{initial_command}{mg_process_directory}/bin/madevent reweight {run_name} -f",
        log_file=log_file,
    )


def copy_ufo_model(ufo_directory, mg_directory):
    _, model_name = os.path.split(ufo_directory)
    destination = f"{mg_directory}/models/{model_name}"

    if os.path.isdir(destination):
        return

    shutil.copytree(ufo_directory, destination)


def create_master_script(log_directory, master_script_filename, mg_directory, mg_process_directory, results):
    placeholder_definition = r"mgdir=${1:-" + mg_directory + r"}" + "\n"
    placeholder_definition += r"mgprocdir=${2:-" + mg_process_directory + r"}" + "\n"
    placeholder_definition += r"mmlogdir=${3:-" + log_directory + r"}"
    commands = "\n".join(results)
    script = (
        "#!/bin/bash\n\n# Master script to generate events for MadMiner\n\n"
        + "# Usage: run.sh [MG_directory] [MG_process_directory] [log_directory]\n\n"
        + "{}\n\n{}"
    ).format(placeholder_definition, commands)

    with open(master_script_filename, "w") as file:
        file.write(script)
    make_file_executable(master_script_filename)
